from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from typing import Dict
import os

DEFAULT_DRIVER_MEMORY = '32000M'
DEFAULT_INSTANCE_TYPE = 'r5.4xlarge'
INSTANCE_DRIVER_MEMORY_MAPPING = {
    'r5.2xlarge': '16000M',
    'r5.xlarge': '8000M',
}


@dataclass
class EmrConfig(BaseConfig):
    ec2_key_name: str = None
    master_security_group: str = None
    slave_security_group: str = None
    master_instance_type: str = DEFAULT_INSTANCE_TYPE
    slave_instance_type: str = DEFAULT_INSTANCE_TYPE

    def get_instances_config(
        self,
        cluster_count: int,
        idle_timeout: int = 0,
        keep_alive: bool = False
    ) -> Dict:
        market = 'ON_DEMAND'
        instances_config = {
            'KeepJobFlowAliveWhenNoSteps': keep_alive,
            'InstanceGroups': [
                dict(
                    Name='AmazonEMRMaster',
                    Market=market,
                    InstanceRole='MASTER',
                    InstanceType=self.master_instance_type,
                    InstanceCount=1,
                    Configurations=[
                        {
                            'Classification': 'spark-defaults',
                            'Properties': {
                                'spark.driver.memory': self.__driver_memory(
                                    self.master_instance_type,
                                ),
                                'spark.driver.maxResultSize': '0',
                                'spark.executor.memory': self.__driver_memory(
                                    self.master_instance_type,
                                ),
                            },
                        },
                    ],
                ),
                dict(
                    Name='AmazonEMRCore',
                    Market=market,
                    InstanceRole='CORE',
                    InstanceType=self.slave_instance_type,
                    InstanceCount=1,
                    Configurations=[
                        {
                            'Classification': 'spark-defaults',
                            'Properties': {
                                'spark.driver.memory': self.__driver_memory(
                                    self.slave_instance_type,
                                ),
                                'spark.driver.maxResultSize': '0',
                                'spark.executor.memory': self.__driver_memory(
                                    self.slave_instance_type,
                                ),
                            },
                        },
                    ],
                ),
            ],
        }
        if os.getenv('MAGE_EC2_SUBNET_ID'):
            instances_config['Ec2SubnetId'] = os.getenv('MAGE_EC2_SUBNET_ID')
        if self.ec2_key_name is not None:
            instances_config['Ec2KeyName'] = self.ec2_key_name
        if self.master_security_group is not None:
            instances_config['EmrManagedMasterSecurityGroup'] = self.master_security_group
        if self.slave_security_group is not None:
            instances_config['EmrManagedSlaveSecurityGroup'] = self.slave_security_group
        return instances_config

    def __driver_memory(self, instance_size: str) -> str:
        return INSTANCE_DRIVER_MEMORY_MAPPING.get(instance_size, DEFAULT_DRIVER_MEMORY)



from google.api_core import retry
from google.cloud import pubsub_v1
from google.oauth2 import service_account
from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.constants import DEFAULT_BATCH_SIZE
from mage_ai.streaming.sources.base import BaseSource
from typing import Callable
import os


@dataclass
class GoogleCloudPubSubConfig(BaseConfig):
    project_id: str
    topic_id: str
    subscription_id: str
    timeout: int = 5
    batch_size: int = DEFAULT_BATCH_SIZE
    pubsub_emulator_host: str = None  # e.g., "host.docker.internal:8085"
    path_to_credentials_json_file: str = None  # e.g., "./google_credentials.json"


class GoogleCloudPubSubSource(BaseSource):
    """
    Handles data transfer between a Google Cloud Pub/Sub topic and the Mage app.

    GOOGLE_APPLICATION_CREDENTIALS environment could be used to set the Google Cloud
    credentials file for authentication.
    """

    config_class = GoogleCloudPubSubConfig

    def _get_publisher_client(self) -> pubsub_v1.PublisherClient:
        if self.config.path_to_credentials_json_file is not None:
            credentials = service_account.Credentials.from_service_account_file(
                self.config.path_to_credentials_json_file
            )
            return pubsub_v1.PublisherClient(credentials=credentials)
        else:
            return pubsub_v1.PublisherClient()

    def _get_subscriber_client(self) -> pubsub_v1.SubscriberClient:
        if self.config.path_to_credentials_json_file is not None:
            credentials = service_account.Credentials.from_service_account_file(
                self.config.path_to_credentials_json_file
            )
            return pubsub_v1.SubscriberClient(credentials=credentials)
        else:
            return pubsub_v1.SubscriberClient()

    def _exist_subscription(self, project_id: str) -> bool:
        project_path = f'projects/{project_id}'
        subscriptions = self.subscriber_client.list_subscriptions(
            request={'project': project_path}
        )
        for subscription in subscriptions:
            if subscription.name == self.subscription_path:
                return True
        return False

    def _create_subscription(
            self, project_id: str, topic_id: str, subscription_id: str) -> None:
        """Create a new pull subscription on the given topic."""
        # [START pubsub_create_pull_subscription]
        if self._exist_subscription(project_id):
            self._print(f'Subscription already exists: {self.subscription_path}')
            return

        publisher = self._get_publisher_client()
        topic_path = publisher.topic_path(project_id, topic_id)
        subscription = self.subscriber_client.create_subscription(
            request={'name': self.subscription_path, 'topic': topic_path}
        )
        self._print(f'Subscription created: {subscription}')
        # [END pubsub_create_pull_subscription]

    def init_client(self) -> None:
        if self.config.pubsub_emulator_host is not None:
            os.environ['PUBSUB_EMULATOR_HOST'] = self.config.pubsub_emulator_host

        self.subscriber_client = self._get_subscriber_client()
        self.subscription_path = self.subscriber_client.subscription_path(
            self.config.project_id, self.config.subscription_id)
        self._create_subscription(
            self.config.project_id,
            self.config.topic_id,
            self.config.subscription_id)

    def read(self, handler: Callable) -> None:
        self._print('Start consuming messages.')

        def callback(
                received_message: pubsub_v1.subscriber.message.Message) -> None:
            # self._print(f'Received: {received_message}.')
            handler(dict(data=received_message.message.data.decode()))
            # self._print(f'Handled: {received_message.message.data}.')
            received_message.ack()

        with self.subscriber_client:
            streaming_pull_future = self.subscriber_client.subscribe(
                self.subscription_path, callback=callback)

            try:
                # When `timeout` is not set, result() will block indefinitely,
                # unless an exception is encountered first.
                self._print('Start receiving message with timeout: {self.config.timeout}')
                streaming_pull_future.result(timeout=self.config.timeout)
            except TimeoutError:
                streaming_pull_future.cancel()  # Trigger the shutdown.
                streaming_pull_future.result()  # Block until the shutdown is complete.

    def batch_read(self, handler: Callable) -> None:
        self._print('Start consuming batch messages.')
        if self.config.batch_size > 0:
            batch_size = self.config.batch_size
        else:
            batch_size = DEFAULT_BATCH_SIZE

        with self.subscriber_client:
            while True:
                response = self.subscriber_client.pull(
                    request={
                        'subscription': self.subscription_path,
                        'max_messages': batch_size
                    },
                    retry=retry.Retry(deadline=300),
                )

                if len(response.received_messages) == 0:
                    continue

                ack_ids = []
                message_values = []
                self._print(f'Number of received messages: '
                            f'{len(response.received_messages)}')
                for received_message in response.received_messages:
                    # self._print(f'Received: {received_message.message.data}.')
                    message_values.append(
                        dict(data=received_message.message.data.decode())
                    )
                    # self._print(f'Batched: {received_message.message.data}.')
                    ack_ids.append(received_message.ack_id)

                handler(message_values)  # Handle the received messages.
                # Acknowledges the received messages so they will not be sent again.
                self.subscriber_client.acknowledge(
                    request={
                        'subscription': self.subscription_path,
                        'ack_ids': ack_ids
                    }
                )

                self._print(
                    f'Received and acknowledged {len(response.received_messages)}'
                    f' messages from {self.subscription_path}.'
                )



class OauthScope():
    SCOPE_DELIMITER = ','

    CLIENT_ALL = 'all'
    CLIENT_INTERNAL = 'internal'
    CLIENT_PRIVATE = 'private'
    CLIENT_PUBLIC = 'public'

    CLIENT_SCOPES = [
        CLIENT_INTERNAL,
        CLIENT_PRIVATE,
        CLIENT_PUBLIC,
    ]

    TOKEN_SCOPES = []



from collections import deque


def build_custom_action_decorator(action_list):
    def custom_action(function):
        action_list.append(function)
        return function

    return custom_action


def execute_custom_action(df, action, **kwargs):
    """
    Handles custom action generation and execution. Custom actions are currently supported in two
    different ways:
    - `transformer_action` decorator: any function decorated with `@transformer_action` will be
    executed afterwards as a custom action. All functions must accept a Pandas dataframe and
    return a Pandas dataframe. Functions decorated with `@transformer_action` are
    executed in the order of their definition.

    - Script style: entering a script into the action code menu will execute the script. The script
     will have access to the dataframe through the `df` variable and must mutate it before
     completion in order to apply the correct action.

    If both custom actions and scripts are provided, the scripts will be executed before the
    custom actions.

    Below is an example of a user defined custom transformer action
    ```python
    @transformer_action
    def transform(df: pandas.DataFrame) -> pandas.DataFrame:
        df['total'] = df.sum(axis=0)
        return df
    ```

    Below is a python script performing the same function that can be run by this function:
    ```python
    df['total'] = df.sum(axis=0)
    ```
    """
    custom_actions = deque()
    transformer_action = build_custom_action_decorator(custom_actions)
    action_code = action['action_code']
    action_code = action_code.replace('\t', ' ' * 4)
    exec(action_code, {'df': df, 'transformer_action': transformer_action})
    while len(custom_actions) != 0:
        df = custom_actions.popleft()(df)
    return df



from mage_ai.orchestration.db.models.oauth import Oauth2AccessToken
from mage_ai.settings import (
    is_disable_pipeline_edit_access,
    DISABLE_NOTEBOOK_EDIT_ACCESS,
    REQUIRE_USER_AUTHENTICATION,
)
from mage_ai.shared.environments import is_test
from typing import Tuple


def authenticate_client_and_token(client_id: str, token: str) -> Tuple[Oauth2AccessToken, bool]:
    oauth_token = Oauth2AccessToken.query.filter(
        Oauth2AccessToken.oauth2_application_id == client_id,
        Oauth2AccessToken.token == token,
    ).first()

    valid = False
    if oauth_token:
        valid = oauth_token.is_valid()

    return oauth_token, valid


def is_owner(user) -> bool:
    return (user and user.owner) or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test())


def has_at_least_admin_role(user) -> bool:
    return not user or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        is_owner(user) or \
        (user.roles and user.roles & 1 != 0)


def has_at_least_editor_role(user) -> bool:
    return not user or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        is_owner(user) or \
        has_at_least_admin_role(user) or \
        (user.roles and user.roles & 2 != 0)


def has_at_least_editor_role_and_notebook_edit_access(user) -> bool:
    return DISABLE_NOTEBOOK_EDIT_ACCESS != 1 and has_at_least_editor_role(user)


def has_at_least_editor_role_and_pipeline_edit_access(user) -> bool:
    return not is_disable_pipeline_edit_access() and has_at_least_editor_role(user)


def has_at_least_viewer_role(user) -> bool:
    return not user or \
        (not REQUIRE_USER_AUTHENTICATION and not is_test()) or \
        is_owner(user) or \
        has_at_least_admin_role(user) or \
        has_at_least_editor_role(user) or \
        (user.roles and user.roles & 4 != 0)



"""Add metrics to block_runs

Revision ID: ec5df57a1c60
Revises: 8971d4cd5b39
Create Date: 2022-12-04 22:18:55.232804

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ec5df57a1c60'
down_revision = '8971d4cd5b39'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('block_run', sa.Column('metrics', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('block_run', 'metrics')
    # ### end Alembic commands ###



from datetime import datetime
from functools import reduce
from mage_ai.api.operations.base import BaseOperation
from mage_ai.api.operations.constants import (
    CREATE,
    DELETE,
    DETAIL,
    LIST,
    UPDATE,
)
from mage_ai.api.logging import debug, error, info
from mage_ai.services.tracking.metrics import increment, timing
from mage_ai.shared.parsers import encode_complex
from typing import Dict, List, Tuple, Union
import json
import simplejson


async def execute_operation(
    handler,
    resource: str,
    pk: Union[None, int, str] = None,
    child: str = None,
    child_pk: Union[None, int, str] = None,
) -> None:
    request = handler.request
    user_id = 'N/A' if request.current_user is None else request.current_user.id

    tags = dict(
        child=child,
        child_pk=child_pk,
        pk=pk,
        resource=resource,
        user_id=user_id,
    )
    start_time = datetime.utcnow()

    if request.error:
        return __render_error(handler, request.error, **tags)

    action, options = __determine_action(
        request, child=child, child_pk=child_pk, pk=pk)
    try:
        response = await BaseOperation(
            action=action,
            files=request.files,
            headers=request.headers,
            meta=__meta(request),
            oauth_client=request.oauth_client,
            oauth_token=request.oauth_token,
            options=options,
            payload=__payload(request),
            pk=child_pk or pk,
            query=__query(request),
            resource=child or resource,
            resource_parent=resource if child else None,
            resource_parent_id=pk if child else None,
            user=request.current_user,
        ).execute()
    except Exception as err:
        __log_error(
            request,
            err,
            **tags,
        )
        raise err

    end_time = datetime.utcnow()

    error = response.get('error', None)
    if error:
        return __render_error(handler, error, **tags)

    info('Action: {} {} {} {} {}'.format(
        action,
        child or resource,
        child_pk or pk,
        resource if child else '',
        pk if child else '',
    ))
    api_time = end_time.timestamp() - start_time.timestamp()
    info(f'Latency: {api_time:.4f} seconds')
    tags = dict(
        action=action,
        resource=child or resource,
        resource_id=child_pk or pk,
        resource_parent=resource if child else None,
        resource_parent_id=pk if child else None,
    )
    timing('api.time', api_time, tags)
    timing('sql.time', api_time, tags)

    handler.write(simplejson.dumps(
        response,
        default=encode_complex,
        ignore_nan=True,
    ))


def __determine_action(
    request,
    child: str = None,
    child_pk: Union[None, int, str] = None,
    pk: Union[None, int, str] = None,
) -> Tuple[str, str]:
    if 'DELETE' == request.method:
        return (DELETE, __parse_request_body(request))

    if 'GET' == request.method:
        if pk and child and child_pk:
            return (DETAIL, request.query_arguments)
        elif pk and child and not child_pk:
            return (LIST, request.query_arguments)
        elif pk and not child:
            return (DETAIL, request.query_arguments)
        elif not pk:
            return (LIST, request.query_arguments)

    if 'POST' == request.method:
        return (CREATE, __parse_request_body(request))

    if 'PUT' == request.method:
        return (UPDATE, __parse_request_body(request))


def __meta(request) -> Dict:
    def _build(obj: Dict, key: str):
        obj[key] = request.query_arguments.get(key)
        return obj

    meta_init = reduce(_build, __meta_keys(request), {})
    meta = {}
    for k, v in meta_init.items():
        if type(v) is not list:
            v = [v]

        arr = []
        for val in v:
            try:
                val = val.decode()
            except (UnicodeDecodeError, AttributeError):
                pass
            arr.append(val)

        meta[k] = arr[0]

    return meta


def __meta_keys(request) -> List[str]:
    return list(
        filter(
            lambda x: x[0] == '_', [
                k for k in request.query_arguments.keys()]))


def __payload(request) -> Dict:
    if 'Content-Type' in request.headers and \
       'multipart/form-data' in request.headers.get('Content-Type'):

        parts = request.body.decode('utf-8', 'ignore').split('\r\n')
        idx = parts.index(
            'Content-Disposition: form-data; name="json_root_body"')
        json_root_body = parts[idx + 2]

        return json.loads(json_root_body)

    return __parse_request_body(request)


def __parse_request_body(request) -> Dict:
    payload = {}

    if request.body_arguments and len(request.body_arguments) >= 1:
        payload = request.body_arguments
    elif request.body:
        if type(request.body) is str:
            payload = json.loads(request.body)
        else:
            payload = json.loads(request.body.decode())

    return payload


def __query(request) -> Dict:
    meta_keys = __meta_keys(request)
    query_arg = dict(request.query_arguments)

    def _build(obj, key):
        value = query_arg.get(key)
        if obj.get(key) and isinstance(obj[key], list):
            obj[key].append(value)
        else:
            obj[key] = value
        return obj
    return reduce(_build, request.query_arguments.keys() - meta_keys, {})


def __render_error(handler, error: Dict, **kwargs):
    __log_error(handler.request, error, **kwargs)
    handler.write(dict(
        error=error,
        status=200,
    ))


def __log_error(
    request,
    err: Dict,
    child: str = None,
    child_pk: str = None,
    pk: str = None,
    resource: str = None,
    user_id: str = None,
):
    if request.current_user is not None:
        error(f'Current user: {request.current_user.id}')
    debug(f'Access token: {request.oauth_token}')

    endpoint = resource
    if pk:
        endpoint += '/{}'.format(pk)
        if child:
            endpoint += '/{}'.format(child)
            if child_pk:
                endpoint += '/{}'.format(child_pk)

    error('[{}] [ERROR] {} /{} [user: {}]: {}'.format(
        datetime.utcnow(),
        request.method,
        endpoint,
        user_id,
        err,
    ))

    increment('api.error', tags=dict(
        endpoint=endpoint,
        error=err['type'] if isinstance(err, dict) else type(err).__name__,
        resource=resource,
    ))



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.KernelPresenter import KernelPresenter


class KernelPolicy(BasePolicy):
    pass


KernelPolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

KernelPolicy.allow_actions([
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())

KernelPolicy.allow_read(KernelPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())


KernelPolicy.allow_read(KernelPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())

KernelPolicy.allow_write([
  'action_type',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())



from http.server import SimpleHTTPRequestHandler
from mage_ai.data_preparation.repo_manager import get_repo_path
import os
import socketserver


def get_dbt_target_path():
    return os.path.join(get_repo_path(), 'dbt/target')


def run_docs_server():
    target_path = get_dbt_target_path()
    if not os.path.exists(target_path):
        os.makedirs(target_path)

    os.chdir(target_path)

    httpd = socketserver.TCPServer(('', 7789), SimpleHTTPRequestHandler)

    try:
        print('Starting DBT docs server...')
        httpd.serve_forever()
    finally:
        httpd.shutdown()
        httpd.server_close()



from mage_ai.api.presenters.BasePresenter import BasePresenter


class FileVersionPresenter(BasePresenter):
    default_attributes = [
        'name',
        'path',
    ]

    def present(self, **kwargs):
        return self.model.to_dict()



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.DIFF

    Calculates difference from previous row along column.

    Docs: https://docs.mage.ai/guides/transformer-blocks#difference
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.DIFF,
        arguments=[],  # Specify at most one column to compute difference with
        axis=Axis.COLUMN,
        outputs=[{'uuid': 'new_diff_column', 'column_type': 'number_with_decimals'}],
    )
{% endblock %}



import json
import os

from google.api.launch_stage_pb2 import LaunchStage
from google.api_core.exceptions import AlreadyExists
from google.cloud import run_v2
from google.oauth2 import service_account

from mage_ai.services.gcp.cloud_run.config import CloudRunConfig


def run_job(command: str, job_id: str, cloud_run_config: CloudRunConfig) -> None:
    if type(cloud_run_config) is dict:
        cloud_run_config = CloudRunConfig.load(config=cloud_run_config)

    scopes = [
        'https://www.googleapis.com/auth/cloud-platform',
        'https://www.googleapis.com/auth/compute',
    ]
    credentials = service_account.Credentials.from_service_account_file(
        cloud_run_config.path_to_credentials_json_file,
        scopes=scopes,
    )

    # Get existing service
    resource_prefix = f'projects/{cloud_run_config.project_id}/locations/{cloud_run_config.region}'
    existing_service_name = os.getenv('GCP_SERVICE_NAME')
    services_client = run_v2.ServicesClient(credentials=credentials)
    existing_service = services_client.get_service(
        run_v2.GetServiceRequest(
            name=f'{resource_prefix}/services/{existing_service_name}'
        )
    )
    service_template = existing_service.template

    # Create job
    jobs_client = run_v2.JobsClient(credentials=credentials)

    containers_with_cmd = service_template.containers
    for c in containers_with_cmd:
        c.command = command.split(' ')
        c.startup_probe = None
    execution_template = run_v2.ExecutionTemplate(
        task_count=1,
        template=run_v2.TaskTemplate(
            containers=containers_with_cmd,
            volumes=service_template.volumes,
            service_account=service_template.service_account,
            execution_environment=service_template.execution_environment,
            encryption_key=service_template.encryption_key,
            vpc_access=service_template.vpc_access
        )
    )
    job = run_v2.Job(
        launch_stage=LaunchStage.BETA,
        template=execution_template,
    )

    request = run_v2.CreateJobRequest(
        parent=resource_prefix,
        job=job,
        job_id=job_id,
    )
    try:
        operation = jobs_client.create_job(request=request)
        print("Waiting for create_job operation to complete...")
        response = operation.result()
        print(json.dumps(response, indent=4, default=str))
    except AlreadyExists:
        pass

    # Run job
    operation = jobs_client.run_job(request=run_v2.RunJobRequest(
        name=f'{resource_prefix}/jobs/{job_id}'
    ))
    print("Waiting for run_job operation to complete...")

    response = operation.result()
    print(json.dumps(response, indent=4, default=str))
    return response



from datetime import datetime, timedelta
from mage_ai.orchestration.db.models.oauth import Oauth2AccessToken, Oauth2Application, User
from typing import Dict
import jwt
import secrets

JWT_ALGORITHM = 'HS256'
JWT_SECRET = 'materia'


def generate_access_token(user: User, application: Oauth2Application = None) -> Oauth2AccessToken:
    token = secrets.token_urlsafe()

    token_count = Oauth2AccessToken.query.filter(Oauth2AccessToken.token == token).count()
    while token_count >= 1:
        token = secrets.token_urlsafe()
        token_count = Oauth2AccessToken.query.filter(Oauth2AccessToken.token == token).count()

    attributes_data = dict(
        expires=datetime.utcnow() + timedelta(days=30),
        token=token,
        user_id=user.id,
    )

    if application:
        attributes_data['oauth2_application_id'] = application.id

    return Oauth2AccessToken.create(**attributes_data)


def get_access_token(token: str) -> Oauth2AccessToken:
    return Oauth2AccessToken.query.filter(Oauth2AccessToken.token == token).first()


def encode_token(token: str, expires: datetime) -> str:
    return jwt.encode({
        'expires': expires.timestamp(),
        'token': token,
    }, JWT_SECRET, algorithm=JWT_ALGORITHM)


def decode_token(encoded_token: str) -> Dict:
    return jwt.decode(encoded_token, key=JWT_SECRET, algorithms=[JWT_ALGORITHM], verify=True)



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.FileContentPresenter import FileContentPresenter


class FileContentPolicy(BasePolicy):
    pass


FileContentPolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

FileContentPolicy.allow_actions([
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

FileContentPolicy.allow_read(FileContentPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
    constants.UPDATE,
])

FileContentPolicy.allow_write([
    'content',
    'name',
    'path',
    'version',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())



from abc import ABC, abstractmethod


class ClusterManager(ABC):
    @abstractmethod
    def create_cluster(self):
        pass

    @abstractmethod
    def list_clusters(self):
        pass

    @abstractmethod
    def set_active_cluster(self, **kwargs):
        pass



from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.variable import Variable, VariableType
from mage_ai.tests.base_test import DBTestCase
from pandas.testing import assert_frame_equal
import numpy as np
import os
import pandas as pd


class VariableTest(DBTestCase):
    def test_write_and_read_data(self):
        pipeline = self.__create_pipeline('test pipeline 1')
        variable1 = Variable('var1', pipeline.dir_path, 'block1')
        variable2 = Variable('var2', pipeline.dir_path, 'block1')
        variable3 = Variable('var3', pipeline.dir_path, 'block2')
        variable4 = Variable('var4', pipeline.dir_path, 'block2')
        variable1.write_data('test')
        variable2.write_data(123)
        variable3.write_data([1, 2, 3, 4])
        variable4.write_data({'k1': 'v1', 'k2': 'v2'})
        self.assertEqual(variable1.read_data(), 'test')
        self.assertEqual(variable2.read_data(), 123)
        self.assertEqual(variable3.read_data(), [1, 2, 3, 4])
        self.assertEqual(variable4.read_data(), {'k1': 'v1', 'k2': 'v2'})

    def test_write_and_read_dataframe(self):
        pipeline = self.__create_pipeline('test pipeline 2')
        variable1 = Variable(
            'var1',
            pipeline.dir_path,
            'block1',
            variable_type=VariableType.DATAFRAME,
        )
        variable2 = Variable('var2', pipeline.dir_path, 'block2')
        df1 = pd.DataFrame(
            [
                [1, 'test'],
                [2, 'test2'],
            ],
            columns=['col1', 'col2']
        )
        df2 = pd.DataFrame(
            [
                [1, 'test', 3.123, np.NaN],
                [2, 'test2', 4.321, np.NaN],
            ],
            columns=['col1', 'col2', 'col3', 'col4']
        )
        df2['col4'] = df2['col4'].astype('Int64')
        variable1.write_data(df1)
        variable2.write_data(df2)
        variable_dir_path = os.path.join(pipeline.dir_path, '.variables')
        self.assertTrue(os.path.exists(
            os.path.join(variable_dir_path, 'block1', 'var1', 'data.parquet'),
        ))
        self.assertTrue(os.path.exists(
            os.path.join(variable_dir_path, 'block1', 'var1', 'sample_data.parquet'),
        ))
        self.assertTrue(os.path.exists(
            os.path.join(variable_dir_path, 'block2', 'var2', 'data.parquet'),
        ))
        self.assertTrue(os.path.exists(
            os.path.join(variable_dir_path, 'block2', 'var2', 'sample_data.parquet'),
        ))
        assert_frame_equal(variable1.read_data(), df1)
        assert_frame_equal(variable1.read_data(sample=True, sample_count=1), df1.iloc[:1])
        assert_frame_equal(variable2.read_data(), df2)
        assert_frame_equal(variable2.read_data(sample=True, sample_count=1), df2.iloc[:1])

    def test_write_and_read_dataframe_analysis(self):
        pipeline = self.__create_pipeline('test pipeline 3')
        variable = Variable(
            'var1',
            pipeline.dir_path,
            'block1',
            variable_type=VariableType.DATAFRAME_ANALYSIS,
        )
        data = dict(
            metadata=dict(
                column_types=dict(
                    col1='number',
                    col2='text',
                ),
            ),
            statistics=dict(
                count=100,
                count_distinct=50,
            ),
            insights=dict(),
            suggestions=[
                dict(
                    title='Remove outliers',
                )
            ]
        )
        variable.write_data(data)
        self.assertEqual(variable.read_data(), data)

    def test_write_and_read_json(self):
        pipeline = self.__create_pipeline('test pipeline 4')
        variable = Variable(
            'var1',
            pipeline.dir_path,
            'block1',
        )
        data = dict(
            results=[100] * 100,
        )
        variable.write_data(data)
        self.assertTrue(os.path.exists(os.path.join(
            self.repo_path,
            'pipelines/test_pipeline_4/.variables/block1/var1/data.json'
        )))
        self.assertTrue(os.path.exists(os.path.join(
            self.repo_path,
            'pipelines/test_pipeline_4/.variables/block1/var1/sample_data.json'
        )))
        self.assertEqual(variable.read_data(), data)
        self.assertEqual(variable.read_data(sample=True), dict(
            results=[100] * 20,
        ))

    def __create_pipeline(self, name):
        pipeline = Pipeline.create(
            name,
            repo_path=self.repo_path,
        )
        block1 = Block.create('block1', 'data_loader', self.repo_path)
        block2 = Block.create('block2', 'transformer', self.repo_path)
        pipeline.add_block(block1)
        pipeline.add_block(block2)
        return pipeline



from typing import Dict
import logging
import sys


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    def __init__(self, logger, log_level=logging.INFO, logging_tags: Dict = dict()):
        self.logger = logger
        self.log_level = log_level
        self.logging_tags = logging_tags or dict()
        self.linebuf = ''
        self.terminal = sys.stdout

    def __getattr__(self, attr):
        return getattr(self.terminal, attr)

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip(), **self.logging_tags)

    def flush(self):
        pass



from dataclasses import dataclass
from datetime import datetime, timezone
from io import BytesIO
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink
from typing import Dict, List
import boto3
import os
import pandas as pd
import sys
import time


@dataclass
class AmazonS3Config(BaseConfig):
    bucket: str
    prefix: str
    file_type: str = 'parquet'
    buffer_size_mb: int = 5
    buffer_timeout_seconds: int = 300
    date_partition_format: str = None


class AmazonS3Sink(BaseSink):
    config_class = AmazonS3Config

    def init_client(self):
        self.client = boto3.client('s3')
        self.last_upload_time = None
        if self.buffer:
            self.upload_data_to_s3()

    def write(self, data: Dict):
        self._print(f'Ingest data {data}, time={time.time()}')
        self.write_buffer([data])

    def batch_write(self, data: List[Dict]):
        if not data:
            return
        self._print(f'Batch ingest {len(data)} records, time={time.time()}. Sample: {data[0]}')
        self.write_buffer(data)

        if self.config.buffer_size_mb and \
                sys.getsizeof(self.buffer) >= self.config.buffer_size_mb * 1_000_000:
            self.upload_data_to_s3()
            return

        if self.config.buffer_timeout_seconds and \
                self.has_buffer_timed_out(self.config.buffer_timeout_seconds):
            self.upload_data_to_s3()
            return

    def upload_data_to_s3(self):
        if not self.buffer:
            return
        self._print(f'Upload {len(self.buffer)} records to S3.')

        df = pd.DataFrame(self.buffer)
        buffer = BytesIO()
        if self.config.file_type == 'parquet':
            df.to_parquet(buffer)
        elif self.config.file_type == 'csv':
            df.to_csv(buffer, index=False)
        else:
            raise Exception(f'File type {self.config.file_type} is not supported.')

        buffer.seek(0)

        curr_time = datetime.now(timezone.utc)

        filename = curr_time.strftime('%Y%m%d-%H%M%S')
        filename = f'{filename}.{self.config.file_type}'

        object_key = self.config.prefix

        date_partition_format = self.config.date_partition_format
        if date_partition_format:
            object_key = os.path.join(object_key, curr_time.strftime(date_partition_format))

        object_key = os.path.join(object_key, filename)

        self.client.put_object(Body=buffer, Bucket=self.config.bucket, Key=object_key)
        self.clear_buffer()



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig


@dataclass
class GoogleChatConfig(BaseConfig):
    webhook_url: str = None

    @property
    def is_valid(self) -> bool:
        return self.webhook_url is not None and self.webhook_url != 'None'



from botocore.config import Config
import boto3
import os
import uuid

EVENT_RULE_LIMIT = 100


def get_all_event_rules():
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    client = boto3.client('events', config=config)
    response = client.list_rules(
        Limit=EVENT_RULE_LIMIT
    )
    formatted_rules = [
        dict(
            name=r['Name'],
            state=r['State'],
            description=r.get('Description'),
            schedule_expression=r.get('ScheduleExpression'),
            event_pattern=r.get('EventPattern'),
        )
        for r in response['Rules']
    ]
    return formatted_rules


def update_event_rule_targets(name):
    lambda_function_arn = os.getenv('LAMBDA_FUNCTION_ARN')
    lambda_function_name = os.getenv('LAMBDA_FUNCTION_NAME')
    if lambda_function_arn is None or lambda_function_name is None:
        return
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    client = boto3.client('events', config=config)
    event_rule_info = client.describe_rule(Name=name)
    targets = client.list_targets_by_rule(Rule=name)['Targets']
    print(f'Current targets for Event rule {name}: {targets}')
    for t in targets:
        if t['Arn'] == lambda_function_arn:
            return
    # Add target to Event rule
    targets.append(dict(Id=str(uuid.uuid4()), Arn=lambda_function_arn))
    response = client.put_targets(Rule=name, Targets=targets)
    print(f'Event rule put_target repsonse: {response}')
    # Add permission to lambda function
    lambda_client = boto3.client('lambda', config=config)
    response = lambda_client.add_permission(
        FunctionName=lambda_function_arn,
        StatementId=str(uuid.uuid4()),
        Action='lambda:InvokeFunction',
        Principal='events.amazonaws.com',
        SourceArn=event_rule_info['Arn'],
    )
    print(f'Lambda function add_permission repsonse: {response}')



from mage_ai.api.presenters.BasePresenter import BasePresenter


class StatusPresenter(BasePresenter):
    default_attributes = [
        'is_instance_manager',
        'repo_path',
        'scheduler_status',
        'instance_type',
        'disable_pipeline_edit_access',
        'require_user_authentication',
    ]



{% extends "data_loaders/default.jinja" %}
{% block imports %}
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.mysql import MySQL
from os import path
{{ super() -}}
{% endblock %}


{% block content %}
@data_loader
def load_data_from_mysql(*args, **kwargs):
    """
    Template for loading data from a MySQL database.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading#mysql
    """
    query = 'Your MySQL query'  # Specify your SQL query here
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    with MySQL.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        return loader.load(query)
{% endblock %}



from mage_ai.data_cleaner.column_types.constants import NUMBER_TYPES

STATUS_NOT_APPLIED = 'not_applied'
STATUS_COMPLETED = 'completed'


class BaseRule:
    default_config = dict()

    def __init__(self, df, column_types, statistics, custom_config={}):
        self.df = df
        self.df_columns = df.columns.tolist()
        self.column_types = column_types
        self.statistics = statistics
        self.custom_config = custom_config

    def _filter_numeric_types(self):
        numeric_columns = []
        numeric_df = self.df.copy()
        for column in self.df_columns:
            if self.column_types[column] in NUMBER_TYPES:
                numeric_df.loc[:, column] = numeric_df.loc[:, column].astype(float)
                numeric_columns.append(column)
            else:
                numeric_df.drop(column, axis=1, inplace=True)
        numeric_df = numeric_df.dropna(axis=0)
        return numeric_df, numeric_columns

    def config(self, name):
        return self.custom_config.get(name, self.default_config.get(name))

    def evaluate(self):
        """Evaluate data cleaning rule and generate suggested actions
        Returns
        -------
        A list of suggested actions
        """
        return []

    def _build_transformer_action_suggestion(
        self,
        title,
        message,
        action_type,
        action_arguments=[],
        action_code='',
        action_options={},
        action_variables={},
        axis='column',
        outputs=[],
    ):
        return dict(
            title=title,
            message=message,
            status=STATUS_NOT_APPLIED,
            action_payload=dict(
                action_type=action_type,
                action_arguments=action_arguments,
                action_code=action_code,
                action_options=action_options,
                action_variables=action_variables,
                axis=axis,
                outputs=outputs,
            ),
        )



from mage_ai.shared.parsers import encode_complex
from mage_ai.server.logger import Logger
import json
import os
import os.path
import pandas as pd
import simplejson

# This is equivalent to ./files
DATA_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), 'files'))

logger = Logger().new_server_logger(__name__)


class Model:
    def __init__(self, id=None, path=None):
        self.path = None
        if path is not None:
            abs_path = os.path.abspath(path)
            self.dir = abs_path
            self.path = abs_path
            return

        # TODO: figure out a good directory to store the files
        if not os.path.isdir(DATA_PATH):
            os.mkdir(DATA_PATH)

        self.path = self.path_name()
        if not os.path.isdir(self.path):
            os.mkdir(self.path)

        if id is None:
            dirs = Model.gen_integer_dir_list(self.path)
            max_id = 0
            if len(dirs) > 0:
                max_id = sorted(dirs, reverse=True)[0]
            self.id = max_id + 1
        else:
            self.id = id

        self.dir = os.path.join(self.path, str(self.id))
        if not os.path.isdir(self.dir):
            os.mkdir(self.dir)

    def read_json_file(self, file_name, default_value={}):
        file_path = os.path.join(self.dir, file_name)
        if not os.path.exists(file_path):
            return default_value
        with open(file_path) as file:
            return json.load(file)

    def write_json_file(self, file_name, obj={}, subdir=None):
        if subdir is None:
            dir_path = self.dir
        else:
            dir_path = os.path.join(self.dir, subdir)
            if not os.path.isdir(dir_path):
                os.mkdir(dir_path)
        with open(os.path.join(dir_path, file_name), 'w') as file:
            simplejson.dump(
                obj,
                file,
                default=encode_complex,
                ignore_nan=True,
            )

    def read_parquet_file(self, file_name):
        file_path = os.path.join(self.dir, file_name)
        if not os.path.exists(file_path):
            return pd.DataFrame()
        return pd.read_parquet(file_path, engine='pyarrow')

    def write_parquet_file(self, file_name, df):
        df_output = df.copy()
        # Clean up data types since parquet doesn't support mixed data types
        for c in df_output.columns:
            series_non_null = df_output[c].dropna()
            if len(series_non_null) > 0:
                coltype = type(series_non_null.iloc[0])
                try:
                    df_output[c] = series_non_null.astype(coltype)
                except Exception:
                    # Fall back to convert to string
                    df_output[c] = series_non_null.astype(str)
        df_output.to_parquet(os.path.join(self.dir, file_name))

    def to_dict(self, detailed):
        pass

    @classmethod
    def folder_name(cls):
        return cls.__name__

    @classmethod
    def is_valid_id(cls, id):
        opts = Model.gen_integer_dir_list(cls.path_name())
        return int(id) in opts

    @classmethod
    def path_name(cls):
        return os.path.join(DATA_PATH, cls.folder_name())

    @classmethod
    def objects(cls):
        arr = []
        dirs = sorted(Model.gen_integer_dir_list(cls.path_name()), reverse=True)
        for id in dirs:
            try:
                arr.append(cls(id=id))
            except Exception:
                logger.exception(f'Fail to load {cls.__name__} with id {id}')
        return arr

    @staticmethod
    def gen_integer_dir_list(pathname):
        dirs = []
        if not os.path.isdir(pathname):
            return []
        for dirname in os.listdir(pathname):
            try:
                dirs.append(int(dirname))
            except ValueError:
                continue
        return dirs



"""Update schedule type enum

Revision ID: 6aecc9bc451c
Revises: 7ac6fed06918
Create Date: 2023-01-05 15:17:52.800183

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6aecc9bc451c'
down_revision = '7ac6fed06918'
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == 'postgresql':
        with op.get_context().autocommit_block():
            op.execute("ALTER TYPE scheduletype ADD VALUE 'API'")
            op.execute("ALTER TYPE scheduletype ADD VALUE 'EVENT'")


def downgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == 'postgresql':
        op.execute("ALTER TYPE scheduletype RENAME TO scheduletype_old")
        op.execute("CREATE TYPE scheduletype AS ENUM('TIME')")
        op.execute((
            "ALTER TABLE pipeline_schedule ALTER COLUMN schedule_type TYPE scheduletype USING "
            "schedule_type::text::scheduletype"
        ))
        op.execute("DROP TYPE scheduletype_old")



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.UserPresenter import UserPresenter


class UserPolicy(BasePolicy):
    def is_current_user(self):
        return self.current_user.owner or self.current_user.id == self.resource.id


UserPolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.is_owner())

UserPolicy.allow_actions([
    constants.DETAIL,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.is_current_user() or policy.has_at_least_admin_role())

UserPolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_admin_role())

UserPolicy.allow_read(UserPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.is_current_user() or policy.has_at_least_admin_role())

UserPolicy.allow_read(UserPresenter.default_attributes + [
    'token',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
], condition=lambda policy: policy.is_current_user() or policy.is_owner())

UserPolicy.allow_write([
    'avatar',
    'email',
    'first_name',
    'last_name',
    'password',
    'password_confirmation',
    'password_current',
    'username',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.is_current_user() or policy.has_at_least_admin_role())

UserPolicy.allow_write([
    'roles',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_admin_role())

UserPolicy.allow_write([
    'owner',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.is_owner())

UserPolicy.allow_write([
    'avatar',
    'email',
    'first_name',
    'last_name',
    'password',
    'password_confirmation',
    'roles',
    'username',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.is_owner())



from datetime import datetime
from distutils.file_util import copy_file
from mage_ai.api.errors import ApiError
from mage_ai.api.utils import (
    authenticate_client_and_token,
    has_at_least_editor_role,
)
from mage_ai.data_preparation.models.constants import (
    BlockType,
    PipelineType,
    CUSTOM_EXECUTION_BLOCK_TYPES,
    PIPELINES_FOLDER,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_config, get_repo_path
from mage_ai.data_preparation.variable_manager import get_global_variables
from mage_ai.orchestration.db.models.oauth import Oauth2Application
from mage_ai.server.active_kernel import (
    get_active_kernel_client,
    get_active_kernel_name,
    switch_active_kernel,
)
from mage_ai.shared.constants import ENV_DEV
from mage_ai.server.logger import Logger
from mage_ai.server.execution_manager import (
    cancel_pipeline_execution,
    check_pipeline_process_status,
    delete_pipeline_copy_config,
    reset_execution_manager,
    set_current_message_task,
    set_current_pipeline_process,
    set_previous_config_path,
)
from mage_ai.server.kernel_output_parser import DataType
from mage_ai.server.kernels import DEFAULT_KERNEL_NAME, KernelName
from mage_ai.server.utils.output_display import (
    add_internal_output_info,
    add_execution_code,
    get_block_output_process_code,
    get_pipeline_execution_code,
)
from mage_ai.settings import (
    is_disable_pipeline_edit_access,
    DISABLE_NOTEBOOK_EDIT_ACCESS,
    HIDE_ENV_VAR_VALUES,
    REQUIRE_USER_AUTHENTICATION,
)
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.security import filter_out_env_var_values
from mage_ai.utils.code import reload_all_repo_modules
from jupyter_client import KernelClient
from typing import Dict, List
import asyncio
import json
import multiprocessing
import os
import re
import tornado.websocket
import traceback
import uuid

logger = Logger().new_server_logger(__name__)


def run_pipeline(
    pipeline: Pipeline,
    config_copy_path: str,
    global_vars: Dict[str, any],
    queue: multiprocessing.Queue,
) -> None:
    """
    Execute pipeline synchronously. This function is meant to be run in a separate process,
    and will write status messages to the passed in multiprocessing queue.
    """
    metadata = dict(
        pipeline_uuid=pipeline.uuid,
    )

    def add_pipeline_message(
        message: str,
        execution_state: str = 'busy',
        metadata: Dict[str, str] = dict(),
        msg_type: str = 'stream_pipeline',
    ):
        msg = dict(
            message=message,
            execution_state=execution_state,
            metadata=metadata,
            msg_type=msg_type,
        )
        queue.put(msg)

    def build_block_output_stdout(
        block_uuid: str,
        execution_state: str = 'busy',
    ):
        return StreamBlockOutputToQueue(
            queue,
            block_uuid,
            execution_state=execution_state,
            metadata=dict(
                block_uuid=block_uuid,
                pipeline_uuid=pipeline.uuid,
            ),
        )

    try:
        pipeline.execute_sync(
            global_vars=global_vars,
            build_block_output_stdout=build_block_output_stdout,
            run_sensors=False,
        )
        add_pipeline_message(
            f'Pipeline {pipeline.uuid} execution complete.\n'
            'You can see the code block output in the corresponding code block.',
            execution_state='idle',
            metadata=metadata,
        )
    except Exception:
        trace = traceback.format_exc().splitlines()
        add_pipeline_message(f'Pipeline {pipeline.uuid} execution failed with error:',
                             metadata=metadata)
        add_pipeline_message(trace, execution_state='idle', metadata=metadata)

    if pipeline.type == PipelineType.PYTHON:
        delete_pipeline_copy_config(config_copy_path)


def publish_pipeline_message(
    message: str,
    execution_state: str = 'busy',
    metadata: Dict[str, str] = dict(),
    msg_type: str = 'stream_pipeline',
) -> None:
    msg_id = str(uuid.uuid4())
    WebSocketServer.send_message(
        dict(
            data=message,
            execution_metadata=metadata,
            execution_state=execution_state,
            msg_id=msg_id,
            msg_type=msg_type,
            type=DataType.TEXT_PLAIN,
        )
    )


class WebSocketServer(tornado.websocket.WebSocketHandler):
    """Simple WebSocket handler to serve clients."""

    # Note that `clients` is a class variable and `send_message` is a
    # classmethod.
    clients = set()
    running_executions_mapping = dict()

    def open(self):
        WebSocketServer.clients.add(self)

    def on_close(self):
        WebSocketServer.clients.remove(self)

    def check_origin(self, origin):
        return True

    def init_kernel_client(self, kernel_name) -> KernelClient:
        if kernel_name != get_active_kernel_name():
            switch_active_kernel(kernel_name)

        return get_active_kernel_client()

    def on_message(self, raw_message):
        message = json.loads(raw_message)

        api_key = message.get('api_key')
        token = message.get('token')

        if REQUIRE_USER_AUTHENTICATION or DISABLE_NOTEBOOK_EDIT_ACCESS:
            valid = not REQUIRE_USER_AUTHENTICATION

            if api_key and token:
                oauth_client = Oauth2Application.query.filter(
                    Oauth2Application.client_id == api_key,
                ).first()
                if oauth_client:
                    oauth_token, valid = authenticate_client_and_token(oauth_client.id, token)
                    valid = valid and \
                        oauth_token and \
                        oauth_token.user and \
                        has_at_least_editor_role(oauth_token.user)
            if not valid or DISABLE_NOTEBOOK_EDIT_ACCESS == 1:
                return self.send_message(
                    dict(
                        data=ApiError.UNAUTHORIZED_ACCESS['message'],
                        execution_metadata=dict(block_uuid=message.get('uuid')),
                        execution_state='idle',
                        msg_id=str(uuid.uuid4()),
                        type=DataType.TEXT_PLAIN,
                    ),
                )

        output = message.get('output')
        if output:
            self.send_message(output)
            return
        global_vars = message.get('global_vars')
        cancel_pipeline = message.get('cancel_pipeline')
        skip_publish_message = message.get('skip_publish_message')
        execute_pipeline = message.get('execute_pipeline')
        check_if_pipeline_running = message.get('check_if_pipeline_running')
        kernel_name = message.get('kernel_name', get_active_kernel_name())
        pipeline_uuid = message.get('pipeline_uuid')
        pipeline = None
        if pipeline_uuid:
            pipeline = Pipeline.get(pipeline_uuid, get_repo_path())

        # Add default trigger runtime variables so the code can run successfully.
        global_vars = {}
        if pipeline_uuid:
            global_vars = message.get(
                'global_vars',
                get_global_variables(pipeline_uuid),
            )
        global_vars['env'] = ENV_DEV
        if 'execution_date' not in global_vars:
            global_vars['execution_date'] = datetime.now()
        global_vars['event'] = dict()

        if cancel_pipeline:
            cancel_pipeline_execution(
                pipeline,
                publish_pipeline_message,
                skip_publish_message,
            )
        elif check_if_pipeline_running:
            check_pipeline_process_status(pipeline, publish_pipeline_message)
        elif not pipeline:
            code = message.get('code')
            # Need to use Python magic command for changing directories
            code = re.sub(r'^!%cd|^!cd', '%cd', code)

            client = self.init_kernel_client(DEFAULT_KERNEL_NAME)
            msg_id = client.execute(code)
            value = dict(block_uuid=message.get('uuid'))
            WebSocketServer.running_executions_mapping[msg_id] = value
        elif execute_pipeline:
            self.__execute_pipeline(
                pipeline,
                kernel_name,
                global_vars,
            )
        else:
            self.__execute_block(
                message,
                pipeline,
                kernel_name,
                global_vars,
            )

    @classmethod
    def send_message(self, message: dict) -> None:
        def should_filter_message(message):
            if message.get('data') is None and message.get('error') is None \
                    and message.get('execution_state') is None and message.get('type') is None:
                return True

            try:
                # Filter out messages meant for jupyter widgets that we can't render
                if message.get('msg_type') == 'display_data' and \
                        message.get('data')[0].startswith('FloatProgress'):
                    return True
            except IndexError:
                pass

            return False

        def filter_out_sensitive_data(message):
            if not message.get('data') or not HIDE_ENV_VAR_VALUES:
                return message
            data = message['data']
            if type(data) is str:
                data = [data]
            data = [filter_out_env_var_values(data_value) for data_value in data]
            message['data'] = data
            return message

        msg_id = message.get('msg_id')
        if msg_id is None:
            return
        if message.get('data') is None and message.get('error') is None \
           and message.get('execution_state') is None and message.get('type') is None:
            return

        if should_filter_message(message):
            return

        message = filter_out_sensitive_data(message)

        execution_metadata = message.get('execution_metadata')
        msg_id_value = execution_metadata if execution_metadata is not None \
            else WebSocketServer.running_executions_mapping.get(msg_id, dict())
        block_type = msg_id_value.get('block_type')
        block_uuid = msg_id_value.get('block_uuid')
        pipeline_uuid = msg_id_value.get('pipeline_uuid')

        error = message.get('error')
        if error:
            message['data'] = self.__format_error(error)

        output_dict = dict(
            block_type=block_type,
            pipeline_uuid=pipeline_uuid,
            uuid=block_uuid,
        )

        message_final = merge_dict(
            message,
            output_dict,
        )

        # KernelResource: when getting a kernel,
        # it will trigger this send_message from the WebSocket subscriber.
        # This log message is unnecessary and publishing to clients is also unnecessary.
        if block_uuid or pipeline_uuid:
            logger.info(
                f'[{block_uuid}] Sending message for {msg_id} to '
                f'{len(self.clients)} client(s):\n{json.dumps(message_final, indent=2)}'
            )

            for client in self.clients:
                client.write_message(json.dumps(message_final))

    def __execute_block(
        self,
        message: Dict[str, any],
        pipeline: Pipeline,
        kernel_name: str,
        global_vars: Dict[str, any],
    ) -> None:
        block_type = message.get('type')
        block_uuid = message.get('uuid')
        custom_code = message.get('code')
        extension_uuid = message.get('extension_uuid')
        run_downstream = message.get('run_downstream')
        run_settings = message.get('run_settings')
        run_tests = message.get('run_tests')
        run_upstream = message.get('run_upstream')
        upstream_blocks = message.get('upstream_blocks')

        pipeline_uuid = pipeline.uuid

        widget = BlockType.CHART == block_type

        block = pipeline.get_block(
            block_uuid,
            block_type=block_type,
            extension_uuid=extension_uuid,
            widget=widget,
        )

        # Execute saved block content when pipeline edits are disabled
        if is_disable_pipeline_edit_access():
            custom_code = block.content

        reload_all_repo_modules(custom_code)

        code = custom_code

        client = self.init_kernel_client(kernel_name)

        value = dict(
            block_type=block_type or block.type,
            block_uuid=block_uuid,
        )

        if not custom_code and BlockType.SCRATCHPAD == block_type:
            self.send_message(
                dict(
                    data='',
                    execution_metadata=value,
                    execution_state='idle',
                    msg_id=str(uuid.uuid4()),
                    type=DataType.TEXT_PLAIN,
                ),
            )
        else:
            if block is not None and block.type in CUSTOM_EXECUTION_BLOCK_TYPES:
                if kernel_name == KernelName.PYSPARK and not widget:
                    remote_execution = True
                else:
                    remote_execution = False
                code = add_execution_code(
                    pipeline_uuid,
                    block_uuid,
                    custom_code,
                    global_vars,
                    block_type=block_type,
                    extension_uuid=extension_uuid,
                    kernel_name=kernel_name,
                    pipeline_config=pipeline.get_config_from_yaml(),
                    repo_config=get_repo_config().to_dict(remote=remote_execution),
                    run_settings=run_settings,
                    run_tests=run_tests,
                    run_upstream=run_upstream,
                    update_status=False if remote_execution else True,
                    upstream_blocks=upstream_blocks,
                    widget=widget,
                )

            msg_id = client.execute(add_internal_output_info(code))

            WebSocketServer.running_executions_mapping[msg_id] = value

            block_output_process_code = get_block_output_process_code(
                pipeline_uuid,
                block_uuid,
                block_type=block_type,
                kernel_name=kernel_name,
            )
            if block_output_process_code is not None:
                client.execute(block_output_process_code)

            if run_downstream:
                # This will only run downstream blocks that are charts/widgets
                for block in block.downstream_blocks:
                    if BlockType.CHART != block.type:
                        continue

                    self.on_message(json.dumps(dict(
                        api_key=message.get('api_key'),
                        code=block.file.content(),
                        pipeline_uuid=pipeline_uuid,
                        token=message.get('token'),
                        type=block.type,
                        uuid=block.uuid,
                    )))

    def __execute_pipeline(
        self,
        pipeline: Pipeline,
        kernel_name: str,
        global_vars: Dict[str, any],
    ) -> None:
        pipeline_uuid = pipeline.uuid

        if kernel_name == KernelName.PYSPARK:
            code = get_pipeline_execution_code(
                pipeline_uuid,
                global_vars=global_vars,
                kernel_name=kernel_name,
                pipeline_config=pipeline.to_dict(include_content=True),
                repo_config=get_repo_config().to_dict(remote=True),
                update_status=False if kernel_name == KernelName.PYSPARK else True,
            )
            client = self.init_kernel_client(kernel_name)
            msg_id = client.execute(code)

            WebSocketServer.running_executions_mapping[msg_id] = dict(
                pipeline_uuid=pipeline_uuid
            )
        else:
            # TODO: save config for other kernel types.
            def save_pipeline_config() -> str:
                pipeline_copy = f'{pipeline.uuid}_{str(uuid.uuid4())}'
                new_pipeline_directory = os.path.join(
                    pipeline.repo_path,
                    PIPELINES_FOLDER,
                    pipeline_copy,
                )
                os.makedirs(new_pipeline_directory, exist_ok=True)
                copy_file(
                    os.path.join(pipeline.dir_path, 'metadata.yaml'),
                    os.path.join(new_pipeline_directory, 'metadata.yaml'),
                )
                set_previous_config_path(new_pipeline_directory)
                return new_pipeline_directory

            reset_execution_manager()

            if pipeline.type == PipelineType.PYTHON:
                publish_pipeline_message(
                    'Saving current pipeline config for backup. This may take some time...',
                    metadata=dict(pipeline_uuid=pipeline_uuid),
                )

                # The pipeline state can potentially break when the execution is cancelled,
                # so we save the pipeline config before execution if the user cancels the
                # excecution.
                config_copy_path = save_pipeline_config()
            else:
                config_copy_path = None

            queue = multiprocessing.Queue()
            proc = multiprocessing.Process(
                target=run_pipeline,
                args=(pipeline, config_copy_path, global_vars, queue)
            )
            proc.start()
            set_current_pipeline_process(proc)

            async def check_for_messages():
                loop = True
                while loop:
                    while not queue.empty():
                        msg = queue.get()
                        metadata = msg.get('metadata')
                        execution_state = msg.get('execution_state')
                        publish_pipeline_message(
                            msg.get('message'),
                            execution_state=execution_state,
                            metadata=metadata,
                            msg_type=msg.get('msg_type'),
                        )
                        if execution_state == 'idle' and \
                                metadata.get('block_uuid') is None:
                            loop = False
                            break
                    await asyncio.sleep(0.5)

            task = asyncio.create_task(check_for_messages())
            set_current_message_task(task)

    def __format_error(error: List[str]) -> List[str]:
        initial_regex = r'.*execute_custom_code\(\).*'
        end_regex = r'.*Block.[_a-z]*\(.*'
        initial_idx = 0
        end_idx = 0
        for idx, line in enumerate(error):
            ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
            line_without_ansi = ansi_escape.sub('', line)
            if re.match(initial_regex, line_without_ansi) and not initial_idx:
                initial_idx = idx
            if re.match(end_regex, line_without_ansi):
                end_idx = idx

        try:
            if initial_idx and end_idx:
                return error[:initial_idx - 1] + error[end_idx + 1:]
        except Exception:
            pass

        return error


class StreamBlockOutputToQueue(object):
    """
    Fake file-like stream object that redirects block output to a queue
    to be streamed to the websocket.
    """
    def __init__(
        self,
        queue,
        block_uuid,
        execution_state='busy',
        metadata=dict(),
        msg_type='stream_pipeline',
    ):
        self.queue = queue
        self.block_uuid = block_uuid
        self.execution_state = execution_state
        self.metadata = metadata
        self.msg_type = msg_type

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.queue.put(dict(
                message=f'[{self.block_uuid}] {line.rstrip()}',
                execution_state=self.execution_state,
                metadata=self.metadata,
                msg_type=self.msg_type,
            ))



from mage_ai.services.opsgenie.config import OpsgenieConfig
import requests
import json


def send_opsgenie_alert(
        config: OpsgenieConfig,
        message: str = None,
        description: str = None
) -> None:
    """
    Opens an alert in Opsgenie with the given message and description.

    Args:
        config (OpsgenieConfig): Opsgenie config dataclass
        message (str): The title of the alert.
        description (str): The message body of the alert.
    """
    url = config.url

    headers = {
        'Content-Type': 'application/json',
        'Authorization': f'GenieKey {config.api_key}'
    }

    payload = {
        'message': message,
        'description': description,
        'priority': config.priority,
        'responders': config.responders,
        'tags': config.tags,
        'details': config.details,
    }

    print(f'Creating Opsgenie Alert {payload}')
    response = requests.post(url, headers=headers, data=json.dumps(payload))

    if response.status_code == 202:
        print(f'Alert successfully opened: {message}')
    else:
        print(f'Unexpected response: {response.status_code} - {response.text}')



from mage_ai.data_preparation.shared.secrets import get_secret_value
from typing import Callable, Dict
import os


def get_template_vars() -> Dict[str, Callable]:
    kwargs = dict(
        env_var=os.getenv,
        mage_secret_var=get_secret_value,
    )

    try:
        from mage_ai.services.aws.secrets_manager.secrets_manager import (
            get_secret
        )
        kwargs['aws_secret_var'] = get_secret
    except ImportError:
        pass

    return kwargs



from mage_ai.streaming.sinks.mongodb import MongoDbSink
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class MongoDbTests(TestCase):
    def test_init(self):
        with patch.object(MongoDbSink, 'init_client') as mock_init_client:
            MongoDbSink(dict(
                connection_string='test_connection_string',
                database_name='test_database_name',
                collection_name='test_collection_name',
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(MongoDbSink, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                MongoDbSink(dict(
                    connection_string='test_connection_string',
                    database_name='test_database_name',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'collection_name\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)



from typing import Dict
from mage_ai.data_preparation.models.constants import PREFERENCES_FILE
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db.models.oauth import User
import os
import traceback
import yaml

# Git environment variables
GIT_REPO_LINK_VAR = 'GIT_REPO_LINK'
GIT_REPO_PATH_VAR = 'GIT_REPO_PATH'
GIT_USERNAME_VAR = 'GIT_USERNAME'
GIT_EMAIL_VAR = 'GIT_EMAIL'
GIT_AUTH_TYPE_VAR = 'GIT_AUTH_TYPE'
GIT_BRANCH_VAR = 'GIT_BRANCH'
GIT_SYNC_ON_PIPELINE_RUN_TYPE = 'GIT_SYNC_ON_PIPELINE_RUN'


class Preferences:
    def __init__(
        self,
        repo_path: str = None,
        config_dict: Dict = None,
        user: User = None,
    ):
        self.repo_path = repo_path or get_repo_path()
        self.preferences_file_path = \
            os.path.join(self.repo_path, PREFERENCES_FILE)
        self.user = user
        preferences = dict()
        try:
            if user:
                preferences = user.preferences or {}
            elif config_dict:
                preferences = config_dict
            elif os.path.exists(self.preferences_file_path):
                with open(self.preferences_file_path) as f:
                    preferences = yaml.full_load(f.read()) or {}
        except Exception:
            traceback.print_exc()
            pass

        if os.getenv(GIT_REPO_LINK_VAR):
            self.sync_config = dict(
                remote_repo_link=os.getenv(GIT_REPO_LINK_VAR),
                repo_path=os.getenv(GIT_REPO_PATH_VAR, os.getcwd()),
                auth_type=os.getenv(GIT_AUTH_TYPE_VAR),
                username=os.getenv(GIT_USERNAME_VAR),
                email=os.getenv(GIT_EMAIL_VAR),
                branch=os.getenv(GIT_BRANCH_VAR),
                sync_on_pipeline_run=bool(int(os.getenv(GIT_SYNC_ON_PIPELINE_RUN_TYPE) or 0)),
            )
        else:
            self.sync_config = preferences.get('sync_config', dict())

    def has_valid_git_config(self) -> bool:
        return 'remote_repo_link' in self.sync_config and 'repo_path' in self.sync_config

    def update_preferences(self, updates: Dict):
        preferences = self.to_dict()
        preferences.update(updates)
        if self.user:
            self.user.update(preferences=preferences)
        else:
            with open(self.preferences_file_path, 'w') as f:
                yaml.dump(preferences, f)

    def to_dict(self) -> Dict:
        return dict(
            sync_config=self.sync_config,
        )


def get_preferences(repo_path=None, user: User = None) -> Preferences:
    default_preferences = Preferences(repo_path=repo_path)
    if user:
        if user.preferences is None \
                and os.path.exists(default_preferences.preferences_file_path):
            return default_preferences
        else:
            return Preferences(user=user)
    else:
        return default_preferences



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.postgres import Postgres
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_postgres(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to a PostgreSQL database.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading#postgresql
    """
    schema_name = 'your_schema_name'  # Specify the name of the schema to export data to
    table_name = 'your_table_name'  # Specify the name of the table to export data to
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    with Postgres.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        loader.export(
            df,
            schema_name,
            table_name,
            index=False,  # Specifies whether to include index in exported table
            if_exists='replace',  # Specify resolution policy if table name already exists
        )



import importlib


class BaseUDF():
    def __init__(self, df, arguments=[], code=None, options={}, kwargs={}):
        self.df = df
        self.arguments = arguments
        self.code = code
        self.options = options
        self.kwargs = kwargs

    def execute(self):
        pass


def execute_udf(udf_name, df, arguments, code, options, kwargs):
    udf_class = getattr(
        importlib.import_module(f'mage_ai.data_cleaner.transformer_actions.udf.{udf_name}'),
        udf_name.title().replace('_', ''),
    )
    return udf_class(df, arguments, code, options, kwargs).execute()



from mage_ai.shared.constants import S3_PREFIX
from urllib.parse import urlparse


def s3_url_path(path):
    if path.startswith(S3_PREFIX):
        s3_url = urlparse(path, allow_fragments=False)
        return s3_url.path.lstrip('/')
    return path



import asyncio
import functools
import json
import os
import sys
import time
import traceback
from contextlib import redirect_stdout
from datetime import datetime
from inspect import Parameter, isfunction, signature
from logging import Logger
from queue import Queue
from typing import Any, Callable, Dict, List, Set, Tuple

import pandas as pd
import simplejson

from mage_ai.data_cleaner.shared.utils import is_geo_dataframe, is_spark_dataframe
from mage_ai.data_preparation.models.block.errors import HasDownstreamDependencies
from mage_ai.data_preparation.models.block.extension.utils import handle_run_tests
from mage_ai.data_preparation.models.block.utils import (
    clean_name,
    fetch_input_variables,
    input_variables,
    is_dynamic_block,
    is_dynamic_block_child,
    is_output_variable,
    is_valid_print_variable,
    output_variables,
    should_reduce_output,
)
from mage_ai.data_preparation.models.constants import (
    BLOCK_LANGUAGE_TO_FILE_EXTENSION,
    CALLBACK_STATUSES,
    CUSTOM_EXECUTION_BLOCK_TYPES,
    DATAFRAME_ANALYSIS_MAX_COLUMNS,
    DATAFRAME_ANALYSIS_MAX_ROWS,
    DATAFRAME_SAMPLE_COUNT_PREVIEW,
    NON_PIPELINE_EXECUTABLE_BLOCK_TYPES,
    BlockColor,
    BlockLanguage,
    BlockStatus,
    BlockType,
    CallbackStatus,
    ExecutorType,
    PipelineType,
)
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.models.variable import VariableType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.shared.stream import StreamToLogger
from mage_ai.data_preparation.templates.template import load_template
from mage_ai.server.kernel_output_parser import DataType
from mage_ai.shared.constants import ENV_DEV, ENV_TEST
from mage_ai.shared.environments import get_env
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.logger import BlockFunctionExec
from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.strings import format_enum
from mage_ai.shared.utils import clean_name as clean_name_orig
from mage_ai.shared.utils import is_spark_env

PYTHON_COMMAND = 'python3'


async def run_blocks(
    root_blocks: List['Block'],
    analyze_outputs: bool = False,
    build_block_output_stdout: Callable[..., object] = None,
    global_vars=None,
    parallel: bool = True,
    run_sensors: bool = True,
    run_tests: bool = True,
    selected_blocks: Set[str] = None,
    update_status: bool = True,
) -> None:
    tries_by_block_uuid = {}
    tasks = dict()
    blocks = Queue()

    def create_block_task(block: 'Block'):
        async def execute_and_run_tests():
            with BlockFunctionExec(
                block.uuid,
                f'Executing {block.type} block...',
                build_block_output_stdout=build_block_output_stdout,
            ):
                await block.execute(
                    analyze_outputs=analyze_outputs,
                    build_block_output_stdout=build_block_output_stdout,
                    global_vars=global_vars,
                    run_all_blocks=True,
                    run_sensors=run_sensors,
                    update_status=update_status,
                    parallel=parallel,
                )

                if run_tests:
                    block.run_tests(
                        build_block_output_stdout=build_block_output_stdout,
                        global_vars=global_vars,
                        update_tests=False,
                    )

        return asyncio.create_task(execute_and_run_tests())

    for block in root_blocks:
        blocks.put(block)
        tasks[block.uuid] = None

    while not blocks.empty():
        block = blocks.get()

        if block.type in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES or \
                not run_sensors and block.type == BlockType.SENSOR:
            continue

        if tries_by_block_uuid.get(block.uuid, None) is None:
            tries_by_block_uuid[block.uuid] = 0

        tries_by_block_uuid[block.uuid] += 1
        tries = tries_by_block_uuid[block.uuid]
        if tries >= 1000:
            raise Exception(f'Block {block.uuid} has tried to execute {tries} times; exiting.')

        skip = False
        for upstream_block in block.upstream_blocks:
            if tasks.get(upstream_block.uuid) is None:
                blocks.put(block)
                skip = True
                break
        if skip:
            continue
        upstream_tasks = [tasks[u.uuid] for u in block.upstream_blocks]
        await asyncio.gather(*upstream_tasks)
        block_task = create_block_task(block)
        tasks[block.uuid] = block_task
        for downstream_block in block.downstream_blocks:
            if downstream_block.uuid not in tasks and (
                selected_blocks is None or upstream_block.uuid in selected_blocks
            ):
                tasks[downstream_block.uuid] = None
                blocks.put(downstream_block)
    remaining_tasks = filter(lambda task: task is not None, tasks.values())
    await asyncio.gather(*remaining_tasks)


def run_blocks_sync(
    root_blocks: List['Block'],
    analyze_outputs: bool = False,
    build_block_output_stdout: Callable[..., object] = None,
    global_vars: Dict = None,
    run_sensors: bool = True,
    run_tests: bool = True,
    selected_blocks: Set[str] = None,
) -> None:
    tries_by_block_uuid = {}
    tasks = dict()
    blocks = Queue()

    for block in root_blocks:
        blocks.put(block)
        tasks[block.uuid] = False

    while not blocks.empty():
        block = blocks.get()

        if block.type in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES:
            continue

        if tries_by_block_uuid.get(block.uuid, None) is None:
            tries_by_block_uuid[block.uuid] = 0

        tries_by_block_uuid[block.uuid] += 1
        tries = tries_by_block_uuid[block.uuid]
        if tries >= 1000:
            raise Exception(f'Block {block.uuid} has tried to execute {tries} times; exiting.')

        skip = False
        for upstream_block in block.upstream_blocks:
            upstream_task_status = tasks.get(upstream_block.uuid)
            if upstream_task_status is None or not upstream_task_status:
                blocks.put(block)
                skip = True
                break
        if skip:
            continue
        with BlockFunctionExec(
            block.uuid,
            f'Executing {block.type} block...',
            build_block_output_stdout=build_block_output_stdout,
        ):
            block.execute_sync(
                analyze_outputs=analyze_outputs,
                build_block_output_stdout=build_block_output_stdout,
                global_vars=global_vars,
                run_all_blocks=True,
                test_execution=not run_sensors,
            )

            if run_tests:
                block.run_tests(
                    build_block_output_stdout=build_block_output_stdout,
                    global_vars=global_vars,
                    update_tests=False,
                )
        tasks[block.uuid] = True
        for downstream_block in block.downstream_blocks:
            if downstream_block.uuid not in tasks and (
                selected_blocks is None or downstream_block.uuid in selected_blocks
            ):
                tasks[downstream_block.uuid] = None
                blocks.put(downstream_block)


class Block:
    def __init__(
        self,
        name: str,
        uuid: str,
        block_type: BlockType,
        block_color: BlockColor = None,
        content: str = None,
        executor_config: Dict = None,
        executor_type: ExecutorType = ExecutorType.LOCAL_PYTHON,
        extension_uuid: str = None,
        status: BlockStatus = BlockStatus.NOT_EXECUTED,
        pipeline=None,
        language: BlockLanguage = BlockLanguage.PYTHON,
        configuration: Dict = dict(),
        has_callback: bool = False,
    ):
        self.name = name or uuid
        self._uuid = uuid
        self.type = block_type
        self._content = content
        self.executor_config = executor_config
        self.executor_type = executor_type
        self.extension_uuid = extension_uuid
        self.status = status
        self.pipeline = pipeline
        self.language = language or BlockLanguage.PYTHON
        self.color = block_color
        self.configuration = configuration
        self.has_callback = has_callback

        self._outputs = None
        self._outputs_loaded = False
        self.callback_blocks = []
        self.upstream_blocks = []
        self.downstream_blocks = []
        self.test_functions = []
        self.global_vars = {}
        self.template_runtime_configuration = {}

        self.dynamic_block_index = None
        self.dynamic_block_uuid = None
        self.dynamic_upstream_block_uuids = None

        # Spark session
        self.spark = None
        self.spark_init = False

    @property
    def uuid(self):
        return self.dynamic_block_uuid or self._uuid

    @uuid.setter
    def uuid(self, x):
        self.dynamic_block_uuid = None
        self._uuid = x

    @property
    def content(self):
        if self._content is None:
            self._content = self.file.content()
        return self._content

    @property
    def callback_block(self):
        if self.has_callback:
            callback_block_uuid = f'{clean_name_orig(self.uuid)}_callback'
            return CallbackBlock(
                callback_block_uuid,
                callback_block_uuid,
                BlockType.CALLBACK,
                pipeline=self.pipeline,
            )
        return None

    async def content_async(self):
        if self._content is None:
            self._content = await self.file.content_async()
        return self._content

    async def metadata_async(self):
        return {}

    @property
    def executable(self):
        return (
            self.type not in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES
            and (self.pipeline is None or self.pipeline.type != PipelineType.STREAMING)
        )

    @property
    def outputs(self):
        if not self._outputs_loaded:
            if self._outputs is None or len(self._outputs) == 0:
                self._outputs = self.get_outputs()
        return self._outputs

    async def outputs_async(self):
        if not self._outputs_loaded:
            if self._outputs is None or len(self._outputs) == 0:
                self._outputs = await self.get_outputs_async()
        return self._outputs

    @property
    def callback_block_uuids(self):
        return [b.uuid for b in self.callback_blocks]

    @property
    def upstream_block_uuids(self):
        return [b.uuid for b in self.upstream_blocks]

    @property
    def downstream_block_uuids(self):
        return [b.uuid for b in self.downstream_blocks]

    @property
    def file_path(self):
        repo_path = self.pipeline.repo_path if self.pipeline is not None else get_repo_path()

        file_extension = BLOCK_LANGUAGE_TO_FILE_EXTENSION[self.language]
        block_directory = f'{self.type}s' if self.type != BlockType.CUSTOM else self.type

        return os.path.join(
            repo_path or os.getcwd(),
            block_directory,
            f'{self.uuid}.{file_extension}',
        )

    @property
    def file(self):
        return File.from_path(self.file_path)

    @property
    def table_name(self) -> str:
        if self.configuration and self.configuration.get('data_provider_table'):
            return self.configuration['data_provider_table']

        table_name = f'{self.pipeline.uuid}_{clean_name_orig(self.uuid)}_'\
                     f'{self.pipeline.version_name}'

        env = (self.global_vars or dict()).get('env')
        if env == ENV_DEV:
            table_name = f'dev_{table_name}'
        elif env == ENV_TEST:
            table_name = f'test_{table_name}'

        return table_name

    @property
    def full_table_name(self) -> str:
        from mage_ai.data_preparation.models.block.sql.utils.shared import (
            extract_create_statement_table_name,
            extract_insert_statement_table_names,
        )

        if not self.content:
            return None

        table_name = extract_create_statement_table_name(self.content)
        if table_name:
            return table_name

        matches = extract_insert_statement_table_names(self.content)
        if len(matches) == 0:
            return None

        return matches[len(matches) - 1]

    @classmethod
    def after_create(self, block: 'Block', **kwargs):
        from mage_ai.data_preparation.models.block.dbt.utils import (
            add_blocks_upstream_from_refs,
        )
        widget = kwargs.get('widget')
        pipeline = kwargs.get('pipeline')
        if pipeline is not None:
            priority = kwargs.get('priority')
            upstream_block_uuids = kwargs.get('upstream_block_uuids', [])

            if BlockType.DBT == block.type and BlockLanguage.SQL == block.language:
                arr = add_blocks_upstream_from_refs(block)
                upstream_block_uuids += [b.uuid for b in arr]
                priority_final = priority if len(upstream_block_uuids) == 0 else None
            else:
                priority_final = priority

            pipeline.add_block(
                block,
                upstream_block_uuids,
                priority=priority_final,
                widget=widget,
            )

    @classmethod
    def block_class_from_type(self, block_type: str, language=None, pipeline=None) -> 'Block':
        from mage_ai.data_preparation.models.block.constants import BLOCK_TYPE_TO_CLASS
        from mage_ai.data_preparation.models.block.dbt import DBTBlock
        from mage_ai.data_preparation.models.block.integration import (
            DestinationBlock,
            SourceBlock,
            TransformerBlock,
        )
        from mage_ai.data_preparation.models.block.r import RBlock
        from mage_ai.data_preparation.models.block.sql import SQLBlock
        from mage_ai.data_preparation.models.widget import Widget

        if BlockType.CHART == block_type:
            return Widget
        elif BlockType.DBT == block_type:
            return DBTBlock
        elif pipeline and PipelineType.INTEGRATION == pipeline.type:
            if BlockType.CALLBACK == block_type:
                return CallbackBlock
            elif BlockType.DATA_LOADER == block_type:
                return SourceBlock
            elif BlockType.DATA_EXPORTER == block_type:
                return DestinationBlock
            else:
                return TransformerBlock
        elif BlockLanguage.SQL == language:
            return SQLBlock
        elif BlockLanguage.R == language:
            return RBlock
        return BLOCK_TYPE_TO_CLASS.get(block_type)

    @classmethod
    def create(
        self,
        name,
        block_type,
        repo_path,
        color=None,
        configuration=None,
        extension_uuid: str = None,
        language=None,
        pipeline=None,
        priority=None,
        upstream_block_uuids=None,
        config=None,
        widget=False,
    ):
        """
        1. Create a new folder for block_type if not exist
        2. Create a new python file with code template
        """
        if upstream_block_uuids is None:
            upstream_block_uuids = []
        if config is None:
            config = {}

        uuid = clean_name(name)
        language = language or BlockLanguage.PYTHON

        if BlockType.DBT != block_type or BlockLanguage.YAML == language:
            block_directory = f'{block_type}s' if block_type != BlockType.CUSTOM else block_type
            block_dir_path = os.path.join(repo_path, block_directory)
            if not os.path.exists(block_dir_path):
                os.mkdir(block_dir_path)
                with open(os.path.join(block_dir_path, '__init__.py'), 'w'):
                    pass

            file_extension = BLOCK_LANGUAGE_TO_FILE_EXTENSION[language]
            file_path = os.path.join(block_dir_path, f'{uuid}.{file_extension}')
            if os.path.exists(file_path):
                if pipeline is not None and pipeline.has_block(
                    uuid,
                    block_type=block_type,
                    extension_uuid=extension_uuid,
                ):
                    raise Exception(f'Block {uuid} already exists. Please use a different name.')
            else:
                load_template(
                    block_type,
                    config,
                    file_path,
                    language=language,
                    pipeline_type=pipeline.type if pipeline is not None else None,
                )

        block = self.block_class_from_type(block_type, pipeline=pipeline)(
            name,
            uuid,
            block_type,
            block_color=color,
            configuration=configuration,
            extension_uuid=extension_uuid,
            language=language,
            pipeline=pipeline,
        )

        if BlockType.DBT == block.type:
            if block.file_path and not block.file.exists():
                block.file.create_parent_directories(block.file_path)
                block.file.update_content('')

        self.after_create(
            block,
            config=config,
            pipeline=pipeline,
            priority=priority,
            upstream_block_uuids=upstream_block_uuids,
            widget=widget,
        )
        return block

    @classmethod
    def get_all_blocks(self, repo_path):
        block_uuids = dict()
        for t in BlockType:
            block_dir = os.path.join(repo_path, f'{t.value}s')
            if not os.path.exists(block_dir):
                continue
            block_uuids[t.value] = []
            for f in os.listdir(block_dir):
                if (f.endswith('.py') or f.endswith('.sql')) and f != '__init__.py':
                    block_uuids[t.value].append(f.split('.')[0])
        return block_uuids

    @classmethod
    def get_block(
        self,
        name,
        uuid,
        block_type,
        configuration=None,
        content=None,
        language=None,
        pipeline=None,
        status=BlockStatus.NOT_EXECUTED,
    ):
        block_class = self.block_class_from_type(
            block_type,
            language=language,
            pipeline=pipeline,
        ) or Block
        return block_class(
            name,
            uuid,
            block_type,
            configuration=configuration,
            content=content,
            language=language,
            pipeline=pipeline,
            status=status,
        )

    def all_upstream_blocks_completed(self, completed_block_uuids: Set[str]):
        return all(b.uuid in completed_block_uuids for b in self.upstream_blocks)

    def delete(
        self,
        widget: bool = False,
        commit: bool = True,
        force: bool = False,
    ) -> None:
        """
        1. If pipeline is not None, delete the block from the pipeline but not delete the block
        file.
        2. If pipeline is None, check whether block is used in any pipelines. If block is being
        used, throw error. Otherwise, delete the block files.
        """
        from mage_ai.data_preparation.models.pipeline import Pipeline

        if self.pipeline is not None:
            self.pipeline.delete_block(
                self,
                widget=widget,
                commit=commit,
                force=force,
            )
            # For block_type SCRATCHPAD and MARKDOWN, also delete the file if possible
            if self.type in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES:
                pipelines = Pipeline.get_pipelines_by_block(self, widget=widget)
                pipelines = [
                    pipeline for pipeline in pipelines if self.pipeline.uuid != pipeline.uuid
                ]
                if len(pipelines) == 0:
                    os.remove(self.file_path)
            return
        # If pipeline is not specified, delete the block from all pipelines and delete the file.
        pipelines = Pipeline.get_pipelines_by_block(self, widget=widget)
        if not force:
            for p in pipelines:
                if not p.block_deletable(self):
                    raise HasDownstreamDependencies(
                        f'Block {self.uuid} has downstream dependencies in pipeline {p.uuid}. '
                        'Please remove the dependencies before deleting the block.'
                    )
        for p in pipelines:
            p.delete_block(
                p.get_block(self.uuid, widget=widget),
                widget=widget,
                commit=commit,
                force=force,
            )
        os.remove(self.file_path)

    def execute_with_callback(
        self,
        global_vars: Dict = None,
        logger: Logger = None,
        logging_tags: Dict = dict(),
        **kwargs
    ):
        """
        This method will execute the block and run the callback functions if they exist
        for this block. This function should only be used when running a block from the
        websocket as a way to test the code in the callback. To run a block in a pipeline
        run, use a BlockExecutor.
        """
        arr = []
        if self.callback_block:
            arr.append(self.callback_block)
        if self.callback_blocks:
            arr += self.callback_blocks

        try:
            output = self.execute_sync(
                global_vars=global_vars,
                logger=logger,
                logging_tags=logging_tags,
                **kwargs
            )
        except Exception as e:
            for callback_block in arr:
                callback_block.execute_callback(
                    'on_failure',
                    global_vars=global_vars,
                    logger=logger,
                    logging_tags=logging_tags,
                )
            raise e

        for callback_block in arr:
            callback_block.execute_callback(
                'on_success',
                global_vars=global_vars,
                logger=logger,
                logging_tags=logging_tags,
            )

        return output

    def execute_sync(
        self,
        analyze_outputs: bool = False,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        execution_partition: str = None,
        global_vars: Dict = None,
        logger: Logger = None,
        logging_tags: Dict = dict(),
        run_all_blocks: bool = False,
        test_execution: bool = False,
        update_status: bool = True,
        store_variables: bool = True,
        verify_output: bool = True,
        input_from_output: Dict = None,
        runtime_arguments: Dict = None,
        dynamic_block_index: int = None,
        dynamic_block_uuid: str = None,
        dynamic_upstream_block_uuids: List[str] = None,
        run_settings: Dict = None,
        **kwargs,
    ) -> Dict:
        try:
            if not run_all_blocks:
                not_executed_upstream_blocks = list(
                    filter(lambda b: b.status == BlockStatus.NOT_EXECUTED, self.upstream_blocks)
                )
                all_upstream_is_dbt = all([BlockType.DBT == b.type
                                           for b in not_executed_upstream_blocks])
                if not all_upstream_is_dbt and len(not_executed_upstream_blocks) > 0:
                    upstream_block_uuids = list(map(lambda b: b.uuid, not_executed_upstream_blocks))
                    raise Exception(
                        f"Block {self.uuid}'s upstream blocks have not been executed yet. "
                        f'Please run upstream blocks {upstream_block_uuids} '
                        'before running the current block.'
                    )
            global_vars = self.__enrich_global_vars(global_vars)

            output = self.execute_block(
                build_block_output_stdout=build_block_output_stdout,
                custom_code=custom_code,
                execution_partition=execution_partition,
                global_vars=global_vars,
                logger=logger,
                logging_tags=logging_tags,
                test_execution=test_execution,
                input_from_output=input_from_output,
                runtime_arguments=runtime_arguments,
                dynamic_block_index=dynamic_block_index,
                dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
                run_settings=run_settings,
                **kwargs,
            )
            block_output = self.post_process_output(output)
            variable_mapping = dict()

            if BlockType.CHART == self.type:
                variable_mapping = block_output
                output = dict(
                    output=simplejson.dumps(
                        block_output,
                        default=encode_complex,
                        ignore_nan=True,
                    )
                )
            else:
                output_count = len(block_output)
                variable_keys = [f'output_{idx}' for idx in range(output_count)]
                variable_mapping = dict(zip(variable_keys, block_output))

            if store_variables and self.pipeline.type != PipelineType.INTEGRATION:
                try:
                    self.store_variables(
                        variable_mapping,
                        execution_partition=execution_partition,
                        override_outputs=True,
                        spark=(global_vars or dict()).get('spark'),
                        dynamic_block_uuid=dynamic_block_uuid,
                    )
                except ValueError as e:
                    if str(e) == 'Circular reference detected':
                        raise ValueError(
                            'Please provide dataframe or json serializable data as output.'
                        )
                    raise e
            # Reset outputs cache
            self._outputs = None

            if update_status:
                self.status = BlockStatus.EXECUTED

            if BlockType.CHART != self.type:
                if analyze_outputs:
                    self.analyze_outputs(variable_mapping)
                else:
                    self.analyze_outputs(variable_mapping, shape_only=True)
        except Exception as err:
            if update_status:
                self.status = BlockStatus.FAILED
            raise err
        finally:
            if update_status:
                self.__update_pipeline_block(widget=BlockType.CHART == self.type)

        return output

    def post_process_output(self, output: Dict) -> List:
        return output['output'] or []

    async def execute(
        self,
        analyze_outputs: bool = False,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        global_vars=None,
        run_all_blocks: bool = False,
        run_sensors: bool = True,
        update_status: bool = True,
        parallel: bool = True,
    ) -> None:
        if parallel:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(
                None,
                functools.partial(
                    self.execute_sync,
                    analyze_outputs=analyze_outputs,
                    build_block_output_stdout=build_block_output_stdout,
                    custom_code=custom_code,
                    global_vars=global_vars,
                    run_all_blocks=run_all_blocks,
                    test_execution=not run_sensors,
                    update_status=update_status,
                )
            )
        else:
            self.execute_sync(
                analyze_outputs=analyze_outputs,
                build_block_output_stdout=build_block_output_stdout,
                custom_code=custom_code,
                global_vars=global_vars,
                run_all_blocks=run_all_blocks,
                test_execution=not run_sensors,
                update_status=update_status,
            )

    def _validate_execution(self, decorated_functions, input_vars):
        """
        Validate whether the number of function arguments matches the upstream blocks.
        Only perform the validation for Python functions.
        """
        if self.type not in CUSTOM_EXECUTION_BLOCK_TYPES:
            return None

        if BlockLanguage.PYTHON != self.language:
            return None

        if len(decorated_functions) == 0:
            raise Exception(
                f'Block {self.uuid} does not have any decorated functions. '
                f'Make sure that a function in the block is decorated with @{self.type}.'
            )
        else:
            block_function = decorated_functions[0]
            sig = signature(block_function)

            num_args = sum(
                arg.kind not in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)
                for arg in sig.parameters.values()
            )
            num_inputs = len(input_vars)
            num_upstream = len(self.upstream_block_uuids)

            has_var_args = num_args != len(sig.parameters)

            if num_args > num_inputs:
                if num_upstream < num_args:
                    raise Exception(
                        f'Block {self.uuid} may be missing upstream dependencies. '
                        f'It expected to have {"at least " if has_var_args else ""}{num_args} '
                        f'arguments, but only received {num_inputs}. '
                        f'Confirm that the @{self.type} method declaration has the correct number '
                        'of arguments.'
                    )
                else:
                    raise Exception(
                        f'Block {self.uuid} is missing input arguments. '
                        f'It expected to have {"at least " if has_var_args else ""}{num_args} '
                        f'arguments, but only received {num_inputs}. '
                        f'Double check the @{self.type} method declaration has the correct number '
                        'of arguments and that the upstream blocks have been executed.'
                    )
            elif num_args < num_inputs and not has_var_args:
                if num_upstream > num_args:
                    raise Exception(
                        f'Block {self.uuid} may have too many upstream dependencies. '
                        f'It expected to have {num_args} arguments, but received {num_inputs}. '
                        f'Confirm that the @{self.type} method declaration has the correct number '
                        'of arguments.'
                    )
                else:
                    raise Exception(
                        f'Block {self.uuid} has too many input arguments. '
                        f'It expected to have {num_args} arguments, but received {num_inputs}. '
                        f'Confirm that the @{self.type} method declaration has the correct number '
                        'of arguments.'
                    )

            return block_function

    def execute_block(
        self,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        execution_partition: str = None,
        input_args: List = None,
        logger: Logger = None,
        logging_tags: Dict = dict(),
        global_vars: Dict = None,
        test_execution: bool = False,
        input_from_output: Dict = None,
        runtime_arguments: Dict = None,
        dynamic_block_index: int = None,
        dynamic_upstream_block_uuids: List[str] = None,
        run_settings: Dict = None,
        **kwargs,
    ) -> Dict:
        # Add pipeline uuid and block uuid to global_vars
        global_vars = merge_dict(
            global_vars or dict(),
            dict(
                pipeline_uuid=self.pipeline.uuid,
                block_uuid=self.uuid,
            ),
        )
        # Set up logger
        if build_block_output_stdout:
            stdout = build_block_output_stdout(self.uuid)
        elif logger is not None:
            stdout = StreamToLogger(logger, logging_tags=logging_tags)
        else:
            stdout = sys.stdout

        # Fetch input variables
        input_vars, kwargs_vars, upstream_block_uuids = self.fetch_input_variables(
            input_args,
            execution_partition,
            global_vars,
            dynamic_block_index=dynamic_block_index,
            dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
        )

        outputs_from_input_vars = {}
        if input_args is None:
            upstream_block_uuids_length = len(upstream_block_uuids)
            for idx, input_var in enumerate(input_vars):
                if idx < upstream_block_uuids_length:
                    upstream_block_uuid = upstream_block_uuids[idx]
                    outputs_from_input_vars[upstream_block_uuid] = input_var
                    outputs_from_input_vars[f'df_{idx + 1}'] = input_var
        else:
            outputs_from_input_vars = dict()

        with redirect_stdout(stdout):
            global_vars_copy = global_vars.copy()
            for kwargs_var in kwargs_vars:
                global_vars_copy.update(kwargs_var)

            outputs = self._execute_block(
                outputs_from_input_vars,
                custom_code=custom_code,
                execution_partition=execution_partition,
                input_vars=input_vars,
                logger=logger,
                logging_tags=logging_tags,
                global_vars=global_vars_copy,
                test_execution=test_execution,
                input_from_output=input_from_output,
                runtime_arguments=runtime_arguments,
                upstream_block_uuids=upstream_block_uuids,
                run_settings=run_settings,
                **kwargs,
            )

        output_message = dict(output=outputs)

        return output_message

    def _execute_block(
        self,
        outputs_from_input_vars,
        custom_code: str = None,
        execution_partition: str = None,
        input_vars: List = None,
        logger: Logger = None,
        logging_tags: Dict = dict(),
        global_vars: Dict = None,
        test_execution: bool = False,
        input_from_output: Dict = None,
        runtime_arguments: Dict = None,
        upstream_block_uuids: List[str] = None,
        run_settings: Dict = None,
        **kwargs,
    ) -> List:
        decorated_functions = []
        test_functions = []

        results = {
            self.type: self._block_decorator(decorated_functions),
            'test': self._block_decorator(test_functions),
        }
        results.update(outputs_from_input_vars)

        if custom_code is not None:
            if BlockType.CHART != self.type or (not self.group_by_columns or not self.metrics):
                exec(custom_code, results)
        elif self.content is not None:
            exec(self.content, results)
        elif os.path.exists(self.file_path):
            with open(self.file_path) as file:
                exec(file.read(), results)

        outputs = []
        if BlockType.CHART == self.type:
            variables = self.get_variables_from_code_execution(results)
            outputs = self.post_process_variables(
                variables,
                code=custom_code,
                results=results,
                upstream_block_uuids=upstream_block_uuids,
            )
        else:
            block_function = self._validate_execution(decorated_functions, input_vars)
            if block_function is not None:
                if logger and 'logger' not in global_vars:
                    global_vars['logger'] = logger
                outputs = self.execute_block_function(
                    block_function,
                    input_vars,
                    global_vars,
                    test_execution,
                )

            if outputs is None:
                outputs = []
            if type(outputs) is not list:
                outputs = [outputs]

            self.test_functions = test_functions

        return outputs

    def execute_block_function(
        self,
        block_function: Callable,
        input_vars: List,
        global_vars: Dict = None,
        test_execution: bool = False,
    ) -> Dict:
        sig = signature(block_function)
        has_kwargs = any([p.kind == p.VAR_KEYWORD for p in sig.parameters.values()])
        if has_kwargs and global_vars is not None and len(global_vars) != 0:
            output = block_function(*input_vars, **global_vars)
        else:
            output = block_function(*input_vars)
        return output

    def exists(self):
        return os.path.exists(self.file_path)

    def fetch_input_variables(
        self,
        input_args,
        execution_partition: str = None,
        global_vars: Dict = None,
        dynamic_block_index: int = None,
        dynamic_upstream_block_uuids: List[str] = None,
    ) -> Tuple[List, List, List]:
        return fetch_input_variables(
            self.pipeline,
            self.upstream_block_uuids,
            input_args,
            execution_partition,
            global_vars,
            dynamic_block_index,
            dynamic_upstream_block_uuids,
        )

    def get_analyses(self):
        if self.status == BlockStatus.NOT_EXECUTED:
            return []
        output_variable_objects = self.output_variable_objects()
        if len(output_variable_objects) == 0:
            return []
        analyses = []
        for v in output_variable_objects:
            if v.variable_type != VariableType.DATAFRAME:
                continue
            data = self.pipeline.variable_manager.get_variable(
                self.pipeline.uuid,
                self.uuid,
                v.uuid,
                variable_type=VariableType.DATAFRAME_ANALYSIS,
            )
            data['variable_uuid'] = v.uuid
            analyses.append(data)
        return analyses

    def get_outputs(
        self,
        execution_partition: str = None,
        include_print_outputs: bool = True,
        sample_count: int = DATAFRAME_SAMPLE_COUNT_PREVIEW,
        variable_type: VariableType = None,
        block_uuid: str = None,
    ) -> List[Dict]:
        if self.pipeline is None:
            return

        if not block_uuid:
            block_uuid = self.uuid

        data_products = []
        outputs = []
        variable_manager = self.pipeline.variable_manager

        all_variables = variable_manager.get_variables_by_block(
            self.pipeline.uuid,
            block_uuid,
            partition=execution_partition,
        )
        if not include_print_outputs:
            all_variables = self.output_variables(execution_partition=execution_partition)

        for v in all_variables:
            variable_object = variable_manager.get_variable_object(
                self.pipeline.uuid,
                block_uuid,
                v,
                partition=execution_partition,
                spark=self.__get_spark_session(),
            )

            if variable_type is not None and variable_object.variable_type != variable_type:
                continue

            data = variable_object.read_data(
                sample=True,
                sample_count=sample_count,
                spark=self.__get_spark_session(),
            )
            if type(data) is pd.DataFrame:
                try:
                    analysis = variable_manager.get_variable(
                        self.pipeline.uuid,
                        block_uuid,
                        v,
                        dataframe_analysis_keys=['metadata', 'statistics'],
                        partition=execution_partition,
                        variable_type=VariableType.DATAFRAME_ANALYSIS,
                    )
                except Exception:
                    analysis = None
                if analysis is not None:
                    stats = analysis.get('statistics', {})
                    column_types = (analysis.get('metadata') or {}).get('column_types', {})
                    row_count = stats.get('original_row_count', stats.get('count'))
                    column_count = stats.get('original_column_count', len(column_types))
                else:
                    row_count, column_count = data.shape

                columns_to_display = data.columns.tolist()[:DATAFRAME_ANALYSIS_MAX_COLUMNS]
                data = dict(
                    sample_data=dict(
                        columns=columns_to_display,
                        rows=json.loads(data[columns_to_display].to_json(orient='split'))['data']
                    ),
                    shape=[row_count, column_count],
                    type=DataType.TABLE,
                    variable_uuid=v,
                )
                data_products.append(data)
                continue
            elif is_geo_dataframe(data):
                data = dict(
                    text_data=f'''Use the code in a scratchpad to get the output of the block:

from mage_ai.data_preparation.variable_manager import get_variable
df = get_variable('{self.pipeline.uuid}', '{self.uuid}', 'df')
''',
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif type(data) is str:
                data = dict(
                    text_data=data,
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif type(data) is dict or type(data) is list:
                data = dict(
                    text_data=simplejson.dumps(
                        data,
                        default=datetime.isoformat,
                        ignore_nan=True,
                    ),
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif is_spark_dataframe(data):
                df = data.toPandas()
                columns_to_display = df.columns.tolist()[:DATAFRAME_ANALYSIS_MAX_COLUMNS]
                data = dict(
                    sample_data=dict(
                        columns=columns_to_display,
                        rows=json.loads(df[columns_to_display].to_json(orient='split'))['data']
                    ),
                    type=DataType.TABLE,
                    variable_uuid=v,
                )
                data_products.append(data)
                continue
            outputs.append(data)
        return outputs + data_products

    async def get_outputs_async(
        self,
        execution_partition: str = None,
        include_print_outputs: bool = True,
        sample_count: int = DATAFRAME_SAMPLE_COUNT_PREVIEW,
        variable_type: VariableType = None,
        block_uuid: str = None,
    ) -> List[Dict]:
        if self.pipeline is None:
            return

        if not block_uuid:
            block_uuid = self.uuid

        data_products = []
        outputs = []
        variable_manager = self.pipeline.variable_manager

        all_variables = variable_manager.get_variables_by_block(
            self.pipeline.uuid,
            block_uuid,
            partition=execution_partition,
        )

        if not include_print_outputs:
            all_variables = self.output_variables(execution_partition=execution_partition)

        for v in all_variables:
            variable_object = variable_manager.get_variable_object(
                self.pipeline.uuid,
                block_uuid,
                v,
                partition=execution_partition,
                spark=self.__get_spark_session(),
            )

            if variable_type is not None and variable_object.variable_type != variable_type:
                continue

            data = await variable_object.read_data_async(
                sample=True,
                sample_count=sample_count,
                spark=self.__get_spark_session(),
            )
            if type(data) is pd.DataFrame:
                try:
                    analysis = variable_manager.get_variable(
                        self.pipeline.uuid,
                        block_uuid,
                        v,
                        dataframe_analysis_keys=['metadata', 'statistics'],
                        partition=execution_partition,
                        variable_type=VariableType.DATAFRAME_ANALYSIS,
                    )
                except Exception:
                    analysis = None
                if analysis is not None:
                    stats = analysis.get('statistics', {})
                    column_types = (analysis.get('metadata') or {}).get('column_types', {})
                    row_count = stats.get('original_row_count', stats.get('count'))
                    column_count = stats.get('original_column_count', len(column_types))
                else:
                    row_count, column_count = data.shape

                columns_to_display = data.columns.tolist()[:DATAFRAME_ANALYSIS_MAX_COLUMNS]
                data = dict(
                    sample_data=dict(
                        columns=columns_to_display,
                        rows=json.loads(data[columns_to_display].to_json(orient='split'))['data']
                    ),
                    shape=[row_count, column_count],
                    type=DataType.TABLE,
                    variable_uuid=v,
                )
                data_products.append(data)
                continue
            elif is_geo_dataframe(data):
                data = dict(
                    text_data=f'''Use the code in a scratchpad to get the output of the block:

from mage_ai.data_preparation.variable_manager import get_variable
df = get_variable('{self.pipeline.uuid}', '{block_uuid}', 'df')
''',
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif type(data) is str:
                data = dict(
                    text_data=data,
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif type(data) is dict or type(data) is list:
                data = dict(
                    text_data=simplejson.dumps(
                        data,
                        default=datetime.isoformat,
                        ignore_nan=True,
                    ),
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif is_spark_dataframe(data):
                df = data.toPandas()
                columns_to_display = df.columns.tolist()[:DATAFRAME_ANALYSIS_MAX_COLUMNS]
                data = dict(
                    sample_data=dict(
                        columns=columns_to_display,
                        rows=json.loads(df[columns_to_display].to_json(orient='split'))['data']
                    ),
                    type=DataType.TABLE,
                    variable_uuid=v,
                )
                data_products.append(data)
                continue
            outputs.append(data)
        return outputs + data_products

    def __save_outputs_prepare(self, outputs):
        variable_mapping = dict()
        for o in outputs:
            if o is None:
                continue
            if all(k in o for k in ['variable_uuid', 'text_data']) and \
                    not is_output_variable(o['variable_uuid']):
                variable_mapping[o['variable_uuid']] = o['text_data']

        self._outputs = outputs
        self._outputs_loaded = True
        return variable_mapping

    def save_outputs(self, outputs, override=False):
        variable_mapping = self.__save_outputs_prepare(outputs)
        self.store_variables(variable_mapping, override=override)

    async def save_outputs_async(self, outputs, override=False):
        variable_mapping = self.__save_outputs_prepare(outputs)
        await self.store_variables_async(variable_mapping, override=override)

    def to_dict_base(self, include_callback_blocks: bool = False):
        language = self.language
        if language and type(self.language) is not str:
            language = self.language.value

        data = dict(
            all_upstream_blocks_executed=all(
                block.status == BlockStatus.EXECUTED for block in self.get_all_upstream_blocks()
            ),
            color=self.color,
            configuration=self.configuration or {},
            downstream_blocks=self.downstream_block_uuids,
            executor_config=self.executor_config,
            executor_type=format_enum(self.executor_type) if self.executor_type else None,
            has_callback=self.has_callback,
            name=self.name,
            language=language,
            status=format_enum(self.status) if self.status else None,
            type=format_enum(self.type) if self.type else None,
            upstream_blocks=self.upstream_block_uuids,
            uuid=self.uuid,
        )

        if include_callback_blocks:
            data['callback_blocks'] = self.callback_block_uuids

        return data

    def to_dict(
        self,
        include_callback_blocks: bool = False,
        include_content: bool = False,
        include_outputs: bool = False,
        sample_count: int = None,
        check_if_file_exists: bool = False,
        **kwargs,
    ):
        data = self.to_dict_base(include_callback_blocks=include_callback_blocks)

        if include_content:
            data['content'] = self.content
            if self.callback_block is not None:
                data['callback_content'] = self.callback_block.content
        if include_outputs:
            data['outputs'] = self.outputs
            if check_if_file_exists:
                file_path = self.file.file_path
                if not os.path.isfile(file_path):
                    data['error'] = dict(
                        error='No such file or directory',
                        message='You may have moved it or changed its filename. '
                        'Delete the current block to remove it from the pipeline or write code ' +
                        f'and save the pipeline to create a new file at {file_path}.',
                    )

        return data

    async def to_dict_async(
        self,
        include_block_metadata: bool = False,
        inclide_block_tags: bool = False,
        include_callback_blocks: bool = False,
        include_content: bool = False,
        include_outputs: bool = False,
        sample_count: int = None,
        check_if_file_exists: bool = False,
    ):
        data = self.to_dict_base(include_callback_blocks=include_callback_blocks)

        if include_content:
            data['content'] = await self.content_async()
            if self.callback_block is not None:
                data['callback_content'] = await self.callback_block.content_async()

        if include_outputs:
            data['outputs'] = await self.outputs_async()
            if check_if_file_exists:
                file_path = self.file.file_path
                if not os.path.isfile(file_path):
                    data['error'] = dict(
                        error='No such file or directory',
                        message='You may have moved it or changed its filename. '
                        'Delete the current block to remove it from the pipeline or write code ' +
                        f'and save the pipeline to create a new file at {file_path}.',
                    )

        if include_block_metadata:
            data['metadata'] = await self.metadata_async()

        if inclide_block_tags:
            data['tags'] = self.tags()

        return data

    def update(self, data, **kwargs):
        if 'name' in data and data['name'] != self.name:
            self.__update_name(data['name'])

        if (
            'type' in data
            and self.type == BlockType.SCRATCHPAD
            and data['type'] != BlockType.SCRATCHPAD
        ):
            self.__update_type(data['type'])

        if 'color' in data and data['color'] != self.color:
            self.color = data['color']
            self.__update_pipeline_block()

        if 'upstream_blocks' in data and set(data['upstream_blocks']) != set(
            self.upstream_block_uuids
        ):
            self.__update_upstream_blocks(data['upstream_blocks'])

        if 'callback_blocks' in data and set(data['callback_blocks']) != set(
            self.callback_block_uuids
        ):
            self.__update_callback_blocks(data['callback_blocks'])

        if 'executor_type' in data and data['executor_type'] != self.executor_type:
            self.executor_type = data['executor_type']
            self.__update_pipeline_block()

        if 'has_callback' in data and data['has_callback'] != self.has_callback:
            self.has_callback = data['has_callback']
            if self.has_callback:
                CallbackBlock.create(self.uuid)
            self.__update_pipeline_block()

        return self

    def update_callback_blocks(self, callback_blocks: List[Any]) -> None:
        self.callback_blocks = callback_blocks

    def update_upstream_blocks(self, upstream_blocks: List[Any]) -> None:
        self.upstream_blocks = upstream_blocks

    def update_content(self, content, widget=False):
        if not self.file.exists():
            raise Exception(f'File for block {self.uuid} does not exist at {self.file.file_path}.')

        if content != self.content:
            self.status = BlockStatus.UPDATED
            self._content = content
            self.file.update_content(content)
            self.__update_pipeline_block(widget=widget)
        return self

    async def update_content_async(self, content, widget=False):
        block_content = await self.content_async()
        if content != block_content:
            self.status = BlockStatus.UPDATED
            self._content = content
            await self.file.update_content_async(content)
            self.__update_pipeline_block(widget=widget)
        return self

    def update_status(self, status: BlockStatus):
        self.status = status
        self.__update_pipeline_block(widget=BlockType.CHART == self.type)

    def get_all_upstream_blocks(self) -> List['Block']:
        queue = Queue()
        visited = set()
        queue.put(self)
        while not queue.empty():
            current_block = queue.get()
            for block in current_block.upstream_blocks:
                if block.uuid == self.uuid:
                    continue
                if block.uuid not in visited:
                    queue.put(block)
                    visited.add(block)
        return list(visited)

    def get_all_downstream_blocks(self) -> List['Block']:
        queue = Queue()
        visited = set()
        queue.put(self)
        while not queue.empty():
            current_block = queue.get()
            for block in current_block.downstream_blocks:
                if block.uuid == self.uuid:
                    continue
                if block.uuid not in visited:
                    queue.put(block)
                    visited.add(block)
        return list(visited)

    def run_upstream_blocks(self) -> None:
        def process_upstream_block(
            block: 'Block',
            root_blocks: List['Block'],
        ) -> List[str]:
            if len(block.upstream_blocks) == 0:
                root_blocks.append(block)
            return block.uuid

        upstream_blocks = self.get_all_upstream_blocks()
        root_blocks = []
        upstream_block_uuids = list(
            map(lambda x: process_upstream_block(x, root_blocks), upstream_blocks)
        )

        run_blocks_sync(root_blocks, selected_blocks=upstream_block_uuids)

    def run_tests(
        self,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        execution_partition: str = None,
        global_vars: Dict = {},
        logger: Logger = None,
        logging_tags: Dict = dict(),
        update_tests: bool = True,
        dynamic_block_uuid: str = None,
        from_notebook: bool = False,
    ) -> None:
        self.dynamic_block_uuid = dynamic_block_uuid

        if self.pipeline \
            and PipelineType.INTEGRATION == self.pipeline.type \
                and self.type in [BlockType.DATA_LOADER, BlockType.DATA_EXPORTER]:

            return

        if build_block_output_stdout:
            stdout = build_block_output_stdout(self.uuid)
        elif logger is not None:
            stdout = StreamToLogger(logger, logging_tags=logging_tags)
        else:
            stdout = sys.stdout

        test_functions = []
        if update_tests:
            results = {
                'test': self._block_decorator(test_functions),
            }
            if custom_code is not None:
                exec(custom_code, results)
            elif os.path.exists(self.file_path):
                with open(self.file_path) as file:
                    exec(file.read(), results)
        else:
            test_functions = self.test_functions

        variable_manager = self.pipeline.variable_manager
        outputs = [
            variable_manager.get_variable(
                self.pipeline.uuid,
                self.uuid,
                variable,
                partition=execution_partition,
            )
            for variable in self.output_variables(execution_partition=execution_partition)
        ]

        with redirect_stdout(stdout):
            tests_passed = 0
            for func in test_functions:
                try:
                    sig = signature(func)
                    has_kwargs = any([p.kind == p.VAR_KEYWORD for p in sig.parameters.values()])
                    if has_kwargs and global_vars is not None and len(global_vars) != 0:
                        func(*outputs, **global_vars)
                    else:
                        func(*outputs)
                    tests_passed += 1
                except AssertionError as err:
                    error_message = f'FAIL: {func.__name__} (block: {self.uuid})'
                    stacktrace = traceback.format_exc()

                    if from_notebook:
                        error_json = json.dumps(dict(
                            error=str(err),
                            message=error_message,
                            stacktrace=stacktrace.split('\n'),
                        ))
                        print(f'[__internal_test__]{error_json}')
                    else:
                        print('==============================================================')
                        print(error_message)
                        print('--------------------------------------------------------------')
                        print(stacktrace)

            message = f'{tests_passed}/{len(test_functions)} tests passed.'
            if from_notebook:
                if len(test_functions) >= 1:
                    success_json = json.dumps(dict(
                        message=message,
                    ))
                    print(f'[__internal_test__]{success_json}')
            else:
                print('--------------------------------------------------------------')
                print(message)
        if tests_passed != len(test_functions):
            raise Exception(f'Failed to pass tests for block {self.uuid}')

        with redirect_stdout(stdout):
            handle_run_tests(
                self,
                dynamic_block_uuid=dynamic_block_uuid,
                execution_partition=execution_partition,
                global_vars=global_vars,
                logger=logger,
                logging_tags=logging_tags,
            )

    def analyze_outputs(self, variable_mapping, shape_only: bool = False):
        if self.pipeline is None:
            return
        for uuid, data in variable_mapping.items():
            if type(data) is pd.DataFrame:
                if data.shape[1] > DATAFRAME_ANALYSIS_MAX_COLUMNS or shape_only:
                    self.pipeline.variable_manager.add_variable(
                        self.pipeline.uuid,
                        self.uuid,
                        uuid,
                        dict(
                            statistics=dict(
                                original_row_count=data.shape[0],
                                original_column_count=data.shape[1],
                            ),
                        ),
                        variable_type=VariableType.DATAFRAME_ANALYSIS,
                    )
                    continue
                if data.shape[0] > DATAFRAME_ANALYSIS_MAX_ROWS:
                    data_for_analysis = data.sample(DATAFRAME_ANALYSIS_MAX_ROWS).reset_index(
                        drop=True,
                    )
                else:
                    data_for_analysis = data.reset_index(drop=True)
                try:
                    from mage_ai.data_cleaner.data_cleaner import clean as clean_data
                    analysis = clean_data(
                        data_for_analysis,
                        df_original=data,
                        transform=False,
                        verbose=False,
                    )
                    self.pipeline.variable_manager.add_variable(
                        self.pipeline.uuid,
                        self.uuid,
                        uuid,
                        dict(
                            metadata=dict(column_types=analysis['column_types']),
                            statistics=analysis['statistics'],
                            insights=analysis['insights'],
                            suggestions=analysis['suggestions'],
                        ),
                        variable_type=VariableType.DATAFRAME_ANALYSIS,
                    )
                except Exception:
                    pass
                    # TODO: we use to silently fail, but it looks bad when using BigQuery
                    # print('\nFailed to analyze dataframe:')
                    # print(traceback.format_exc())

    def set_global_vars(self, global_vars):
        self.global_vars = global_vars
        for upstream_block in self.upstream_blocks:
            upstream_block.global_vars = global_vars

    def __consolidate_variables(self, variable_mapping: Dict):
        # Consolidate print variables
        output_variables = {k: v for k, v in variable_mapping.items() if is_output_variable(k)}
        print_variables = {k: v for k, v in variable_mapping.items()
                           if is_valid_print_variable(k, v, self.uuid)}

        print_variables_keys = sorted(print_variables.keys(), key=lambda k: int(k.split('_')[-1]))

        consolidated_print_variables = dict()
        state = dict(
            msg_key=None,
            msg_value=None,
            msg_type=None,
            consolidated_data=None,
        )

        def save_variable_and_reset_state():
            if state['msg_key'] is not None and state['msg_value'] is not None:
                state['msg_value']['data'] = state['consolidated_data']
                consolidated_print_variables[state['msg_key']] = json.dumps(state['msg_value'])
            state['msg_key'] = None
            state['msg_value'] = None
            state['msg_type'] = None
            state['consolidated_data'] = None

        for k in print_variables_keys:
            value = print_variables[k]
            try:
                json_value = json.loads(value)
            except Exception:
                consolidated_print_variables[k] = value
                save_variable_and_reset_state()
                continue

            if 'msg_type' not in json_value or 'data' not in json_value:
                consolidated_print_variables[k] = value
                save_variable_and_reset_state()
                continue

            if state['msg_key'] is not None and json_value['msg_type'] != state['msg_type']:
                save_variable_and_reset_state()

            data = json_value['data'] if type(json_value['data']) is list else [json_value['data']]
            if state['msg_key'] is None:
                state['msg_key'] = k
                state['msg_type'] = json_value['msg_type']
                state['msg_value'] = json_value
                state['consolidated_data'] = data
            else:
                if state['consolidated_data'][-1] == '':
                    # New line
                    state['consolidated_data'] = state['consolidated_data'][:-1] + data
                else:
                    # Not new line
                    state['consolidated_data'][-1] += data[0]
                    state['consolidated_data'] += data[1:]
        save_variable_and_reset_state()

        variable_mapping = merge_dict(output_variables, consolidated_print_variables)
        return variable_mapping

    def __enrich_global_vars(self, global_vars: Dict = None):
        global_vars = global_vars or dict()
        if ((self.pipeline is not None and self.pipeline.type == PipelineType.DATABRICKS) or
                is_spark_env()):
            if not global_vars.get('spark'):
                spark = self.__get_spark_session()
                if spark is not None:
                    global_vars['spark'] = spark
        if 'env' not in global_vars:
            global_vars['env'] = get_env()
        if 'configuration' not in global_vars:
            global_vars['configuration'] = self.configuration
        return global_vars

    def __get_spark_session(self):
        if self.spark_init:
            return self.spark
        try:
            from pyspark.sql import SparkSession
            self.spark = SparkSession.builder.master(
                os.getenv('SPARK_MASTER_HOST', 'local')).getOrCreate()
        except Exception:
            self.spark = None
        self.spark_init = True
        return self.spark

    def __store_variables_prepare(
        self,
        variable_mapping: Dict,
        execution_partition: str = None,
        override: bool = False,
        override_outputs: bool = False,
        dynamic_block_uuid: str = None,
    ):
        self.dynamic_block_uuid = dynamic_block_uuid

        if self.pipeline is None:
            return
        all_variables = self.pipeline.variable_manager.get_variables_by_block(
            self.pipeline.uuid,
            self.uuid,
            partition=execution_partition,
        )

        variable_mapping = self.__consolidate_variables(variable_mapping)

        variable_names = [clean_name_orig(v) for v in variable_mapping]
        removed_variables = []
        for v in all_variables:
            if v in variable_names:
                continue

            is_output_var = is_output_variable(v)
            if (override and not is_output_var) or (override_outputs and is_output_var):
                removed_variables.append(v)
        return dict(
            removed_variables=removed_variables,
            variable_mapping=variable_mapping,
        )

    def store_variables(
        self,
        variable_mapping: Dict,
        execution_partition: str = None,
        override: bool = False,
        override_outputs: bool = False,
        spark=None,
        dynamic_block_uuid: str = None,
    ):
        variables_data = self.__store_variables_prepare(
            variable_mapping,
            execution_partition,
            override,
            override_outputs,
            dynamic_block_uuid,
        )
        for uuid, data in variables_data['variable_mapping'].items():
            if spark is not None and self.pipeline.type == PipelineType.PYSPARK \
                    and type(data) is pd.DataFrame:
                data = spark.createDataFrame(data)
            self.pipeline.variable_manager.add_variable(
                self.pipeline.uuid,
                self.uuid,
                uuid,
                data,
                partition=execution_partition,
            )

        for uuid in variables_data['removed_variables']:
            self.pipeline.variable_manager.delete_variable(
                self.pipeline.uuid,
                self.uuid,
                uuid,
            )

    async def store_variables_async(
        self,
        variable_mapping: Dict,
        execution_partition: str = None,
        override: bool = False,
        override_outputs: bool = False,
        spark=None,
        dynamic_block_uuid: str = None,
    ):
        variables_data = self.__store_variables_prepare(
            variable_mapping,
            execution_partition,
            override,
            override_outputs,
            dynamic_block_uuid,
        )
        for uuid, data in variables_data['variable_mapping'].items():
            if spark is not None and type(data) is pd.DataFrame:
                data = spark.createDataFrame(data)
            await self.pipeline.variable_manager.add_variable_async(
                self.pipeline.uuid,
                self.uuid,
                uuid,
                data,
                partition=execution_partition,
            )

        for uuid in variables_data['removed_variables']:
            self.pipeline.variable_manager.delete_variable(
                self.pipeline.uuid,
                self.uuid,
                uuid,
            )

    def input_variables(self, execution_partition: str = None) -> Dict[str, List[str]]:
        """Get input variables from upstream blocks' output variables.
        Args:
            execution_partition (str, optional): The execution paratition string.

        Returns:
            Dict[str, List[str]]: Mapping from upstream block uuid to a list of variable names
        """
        return input_variables(self.pipeline, self.upstream_block_uuids, execution_partition)

    def input_variable_objects(self, execution_partition: str = None) -> List:
        """Get input variable objects from upstream blocks' output variables.

        Args:
            execution_partition (str, optional): The execution paratition string.

        Returns:
            List: List of input variable objects.
        """
        objs = []
        for b in self.upstream_blocks:
            for v in b.output_variables(execution_partition=execution_partition):
                objs.append(
                    self.pipeline.variable_manager.get_variable_object(
                        self.pipeline.uuid,
                        b.uuid,
                        v,
                        partition=execution_partition,
                    ),
                )
        return objs

    def output_variables(self, execution_partition: str = None) -> List[str]:
        return output_variables(
            self.pipeline,
            self.uuid,
            execution_partition,
        )

    def output_variable_objects(
        self,
        execution_partition: str = None,
        variable_type: VariableType = None,
    ) -> List:
        """Get output variable objects.

        Args:
            execution_partition (str, optional): The execution paratition string.

        Returns:
            List: List of output variable objects.
        """
        if self.pipeline is None:
            return []

        output_variables = self.output_variables(execution_partition=execution_partition)

        if len(output_variables) == 0:
            return []

        variable_objects = [self.pipeline.variable_manager.get_variable_object(
            self.pipeline.uuid,
            self.uuid,
            v,
            partition=execution_partition,
        ) for v in output_variables]
        if variable_type is not None:
            variable_objects = [v for v in variable_objects if v.variable_type == variable_type]
        return variable_objects

    def tags(self) -> List[str]:
        from mage_ai.data_preparation.models.block.constants import (
            TAG_DYNAMIC,
            TAG_DYNAMIC_CHILD,
            TAG_REDUCE_OUTPUT,
        )

        arr = []

        if is_dynamic_block(self):
            arr.append(TAG_DYNAMIC)

        if is_dynamic_block_child(self):
            arr.append(TAG_DYNAMIC_CHILD)

        if should_reduce_output(self):
            arr.append(TAG_REDUCE_OUTPUT)

        return arr

    def variable_object(
        self,
        variable_uuid: str,
        execution_partition: str = None,
    ):
        if self.pipeline is None:
            return []
        return self.pipeline.variable_manager.get_variable_object(
            self.pipeline.uuid,
            self.uuid,
            variable_uuid,
            partition=execution_partition,
        )

    def _block_decorator(self, decorated_functions):
        def custom_code(function):
            decorated_functions.append(function)
            return function

        return custom_code

    # TODO: Update all pipelines that use this block
    def __update_name(self, name):
        """
        1. Rename block file
        2. Update the folder of variable
        3. Update upstream and downstream relationships
        """
        old_uuid = self.uuid
        old_file_path = self.file_path
        new_uuid = clean_name(name)
        self.name = name
        self.uuid = new_uuid
        new_file_path = self.file_path
        if self.pipeline is not None:
            if self.pipeline.has_block(
                new_uuid,
                block_type=self.type,
                extension_uuid=self.extension_uuid,
            ):
                raise Exception(
                    f'Block {new_uuid} already exists in pipeline. Please use a different name.'
                )

        if os.path.exists(new_file_path):
            raise Exception(f'Block {new_uuid} already exists. Please use a different name.')

        file_path_parts = new_file_path.split(os.sep)
        parent_dir = os.path.join(*file_path_parts[:-1])
        os.makedirs(parent_dir, exist_ok=True)

        os.rename(old_file_path, new_file_path)
        if self.pipeline is not None:
            self.pipeline.update_block_uuid(self, old_uuid, widget=BlockType.CHART == self.type)

    def __update_pipeline_block(self, widget=False):
        if self.pipeline is None:
            return
        self.pipeline.update_block(self, widget=widget)

    def __update_type(self, block_type):
        """
        1. Move block file to another folder
        2. Update the block type in pipeline metadata.yaml
        3. Update the code in block file
        """
        old_file_path = self.file_path
        self.type = block_type
        new_file_path = self.file_path
        if os.path.exists(new_file_path):
            raise Exception(
                f'Block {self.type}{os.sep}{self.uuid} already exists.'
                ' Please rename it before changing the type.'
            )
        os.rename(old_file_path, new_file_path)
        if self.pipeline is not None:
            self.pipeline.update_block(self)
        with open(new_file_path) as f:
            existing_code = f.read()
        load_template(
            block_type,
            dict(existing_code='    ' + existing_code.replace('\n', '\n    ')),
            new_file_path,
            language=self.language,
        )

    def __update_upstream_blocks(self, upstream_blocks):
        if self.pipeline is None:
            return
        self.pipeline.update_block(
            self,
            upstream_block_uuids=upstream_blocks,
            widget=BlockType.CHART == self.type,
        )

    def __update_callback_blocks(self, block_uuids: List[str]):
        if self.pipeline is None:
            return

        self.pipeline.update_block(
            self,
            callback_block_uuids=block_uuids,
            widget=BlockType.CHART == self.type,
        )


class SensorBlock(Block):
    def execute_block_function(
        self,
        block_function: Callable,
        input_vars: List,
        global_vars: Dict = None,
        test_execution: bool = False,
    ) -> List:
        if test_execution:
            return super().execute_block_function(
                block_function,
                input_vars,
                global_vars=global_vars,
                test_execution=test_execution,
            )
        else:
            sig = signature(block_function)
            has_kwargs = any([p.kind == p.VAR_KEYWORD for p in sig.parameters.values()])
            use_global_vars = has_kwargs and global_vars is not None and len(global_vars) != 0
            while True:
                condition = block_function(**global_vars) if use_global_vars else block_function()
                if condition:
                    break
                print('Sensor sleeping for 1 minute...')
                time.sleep(60)
            return []


class CallbackBlock(Block):
    @classmethod
    def create(cls, orig_block_name):
        return Block.create(
            f'{clean_name_orig(orig_block_name)}_callback',
            BlockType.CALLBACK,
            get_repo_path(),
            language=BlockLanguage.PYTHON,
        )

    def execute_callback(
        self,
        callback: str,
        global_vars: Dict = None,
        logger: Logger = None,
        logging_tags: Dict = None,
        parent_block: Block = None,
        **kwargs
    ):
        pipeline_run = kwargs.get('pipeline_run')

        if logger is not None:
            stdout = StreamToLogger(logger, logging_tags=logging_tags)
        else:
            stdout = sys.stdout

        with redirect_stdout(stdout):
            global_vars = merge_dict(
                global_vars or dict(),
                dict(
                    pipeline_uuid=self.pipeline.uuid,
                    block_uuid=self.uuid,
                    pipeline_run=pipeline_run,
                ),
            )

            if parent_block and \
                    parent_block.pipeline and \
                    PipelineType.INTEGRATION == parent_block.pipeline.type:

                template_runtime_configuration = parent_block.template_runtime_configuration
                index = template_runtime_configuration.get('index', None)
                is_last_block_run = template_runtime_configuration.get('is_last_block_run', False)
                selected_streams = template_runtime_configuration.get('selected_streams', [])
                stream = selected_streams[0] if len(selected_streams) >= 1 else None
                destination_table = template_runtime_configuration.get('destination_table', stream)

                global_vars['index'] = index
                global_vars['is_last_block_run'] = is_last_block_run
                global_vars['stream'] = stream
                global_vars['destination_table'] = destination_table

            callback_functions = []
            failure_functions = []
            success_functions = []

            results = dict(
                callback=self._block_decorator(callback_functions),
                on_failure=super()._block_decorator(failure_functions),
                on_success=super()._block_decorator(success_functions),
            )
            exec(self.content, results)

            callback_functions_legacy = []
            callback_status = None

            if 'on_failure' == callback:
                callback_functions_legacy = failure_functions
                callback_status = CallbackStatus.FAILURE
            elif 'on_success' == callback:
                callback_functions_legacy = success_functions
                callback_status = CallbackStatus.SUCCESS

            for callback_function in callback_functions_legacy:
                callback_function(**global_vars)

            for callback_function in callback_functions:
                callback_function(callback_status, **global_vars)

    def update_content(self, content, widget=False):
        if not self.file.exists():
            raise Exception(f'File for block {self.uuid} does not exist at {self.file.file_path}.')

        if content != self.content:
            self._content = content
            self.file.update_content(content)
        return self

    async def update_content_async(self, content, widget=False):
        block_content = await self.content_async()
        if content != block_content:
            self._content = content
            await self.file.update_content_async(content)
        return self

    def _block_decorator(self, decorated_functions):
        def custom_code(callback_status: CallbackStatus = CallbackStatus.SUCCESS, *args, **kwargs):
            # If the decorator is just @callback with no arguments, default to success callback
            if isfunction(callback_status):
                def func(callback_status_inner, *args, **kwargs):
                    if callback_status_inner == CallbackStatus.SUCCESS:
                        return callback_status(*args, **kwargs)
                decorated_functions.append(func)
                return func

            if callback_status not in CALLBACK_STATUSES:
                raise Exception(
                    f"Callback status '{callback_status}' in @callback decorator must be 1 of: "
                    f"{', '.join(CALLBACK_STATUSES)}",
                )

            def inner(function):
                def func(callback_status_inner, *args, **kwargs):
                    if callback_status_inner == callback_status:
                        return function(*args, **kwargs)
                decorated_functions.append(func)

            return inner

        return custom_code



{% extends 'data_loaders/deltalake/default.jinja' %}
{% block storage_options %}
    storage_options = {
        'AWS_ACCESS_KEY_ID': '',
        'AWS_SECRET_ACCESS_KEY': '',
        'AWS_REGION': '',
        'AWS_S3_ALLOW_UNSAFE_RENAME': 'true',
    }
{% endblock %}

{% block uri %}
    uri = 's3://[bucket]/[key]'
{% endblock %}



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.file import File
from mage_ai.orchestration.db import safe_db_query


class FileVersionResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        parent_model = kwargs.get('parent_model')

        if isinstance(parent_model, Block):
            pass
        elif isinstance(parent_model, File):
            return self.build_result_set(
                parent_model.file_versions(),
                user,
                **kwargs,
            )
        else:
            raise ApiError(ApiError.RESOURCE_INVALID.copy())



from typing import Dict

from mage_ai.data_preparation.models.block.sql.utils.shared import (
    create_upstream_block_tables as create_upstream_block_tables_orig,
)
from mage_ai.data_preparation.models.block.sql.utils.shared import interpolate_input
from mage_ai.io.config import ConfigKey


def create_upstream_block_tables(
    loader,
    block,
    cascade_on_drop: bool = False,
    configuration: Dict = None,
    execution_partition: str = None,
    cache_upstream_dbt_models: bool = False,
    query: str = None,
):
    create_upstream_block_tables_orig(
        loader,
        block,
        cascade_on_drop,
        configuration,
        execution_partition,
        cache_upstream_dbt_models,
        cache_keys=[
            ConfigKey.POSTGRES_DBNAME,
            ConfigKey.POSTGRES_HOST,
            ConfigKey.POSTGRES_PORT,
            # ConfigKey.POSTGRES_SCHEMA,
        ],
        query=query,
        schema_name=loader.default_schema(),
    )


def interpolate_input_data(block, query, loader):
    return interpolate_input(
        block,
        query,
        get_database=lambda opts: f'"{loader.default_database()}"',
        get_schema=lambda opts: loader.default_schema(),
    )



from mage_ai.api.constants import META_KEY_LIMIT, META_KEY_OFFSET
from typing import Dict, List


def limit_collection(results: List, query: Dict) -> List:
    limit = query.get(META_KEY_LIMIT, [None])
    if limit:
        limit = limit[0]

    offset = query.get(META_KEY_OFFSET, [None])
    if offset:
        offset = offset[0]

    if limit is not None:
        results = results.limit(limit)

    if offset is not None:
        results = results.offset(offset)

    return results



from abc import ABC, abstractmethod
from enum import Enum
from typing import Callable, Dict
import json


class SourceConsumeMethod(str, Enum):
    BATCH_READ = 'BATCH_READ'
    READ = 'READ'
    READ_ASYNC = 'READ_ASYNC'


class BaseSource(ABC):
    config_class = None
    consume_method = SourceConsumeMethod.BATCH_READ

    def __init__(self, config: Dict, **kwargs):
        if self.config_class is not None:
            if 'connector_type' in config:
                config.pop('connector_type')
            self.config = self.config_class.load(config=config)
        self.checkpoint_path = kwargs.get('checkpoint_path')
        self.checkpoint = self.read_checkpoint()
        self.init_client()

    def init_client():
        pass

    @abstractmethod
    def read(self, handler: Callable):
        pass

    async def read_async(self, handler: Callable):
        self._print('Start consuming messages asynchronously.')
        return self.read(handler)

    @abstractmethod
    def batch_read(self, handler: Callable):
        pass

    def read_checkpoint(self):
        checkpoint = None
        if self.checkpoint_path is None:
            return None
        try:
            with open(self.checkpoint_path) as fp:
                checkpoint = json.load(fp)
        except Exception:
            pass
        return checkpoint

    def update_checkpoint(self):
        if self.checkpoint_path is None or self.checkpoint is None:
            return
        try:
            with open(self.checkpoint_path, 'w') as fp:
                json.dump(self.checkpoint, fp)
        except Exception:
            pass

    def test_connection(self):
        return True

    def _print(self, msg):
        print(f'[{self.__class__.__name__}] {msg}')



from datadog import initialize, api, statsd
from datetime import datetime
import os
import time

options = {
    'api_key': os.getenv('DD_API_KEY'),
    'app_key': os.getenv('DD_APP_KEY'),
}

initialize(**options)


def create_event(title, text, tags={}):
    return api.Event.create(
        title=title,
        text=text,
        tags=tags,
    )


def gauge(metric, value, host='mage', tags={}):
    return api.Metric.send(
        host=host,
        metric=metric,
        points=[
            (datetime.utcnow().timestamp(), value),
        ],
        tags=tags,
        type='gauge',
    )


def increment(metric, tags={}, value=1):
    return create_metrics([
        (metric, value, tags),
    ])


def create_metrics(metrics, host='mage', metric_type='count'):
    # Format of argument
    # metrics = [
    #     ('metric', 'value', 'tags'),
    # ]
    arr = [{
        'host': host,
        'metric': t[0],
        'points': t[1],
        'tags': t[2] if len(t) == 3 else {},
        'type': metric_type,
    } for t in metrics]
    return api.Metric.send(metrics=arr)


def create_metric(metric, value, host='mage', tags={}):
    return api.Metric.send(
        host=host,
        metric=metric,
        points=[
            (datetime.utcnow().timestamp(), value),
        ],
        tags=tags,
        type='count',
    )


def histogram(metric, value, host='mage', tags={}):
    return api.Metric.send(
        host=host,
        metric=metric,
        points=[
            (datetime.utcnow().timestamp(), value),
        ],
        tags=tags,
        type='histogram',
    )


def timing(metric, value, host='mage', tags={}):
    return api.Metric.send(
        host=host,
        metric=metric,
        points=[
            (datetime.utcnow().timestamp(), value),
        ],
        tags=tags,
        type='timer',
    )


class timed_decorator(object):
    """
    @timed_decorator('metric.metric', tags={ 'key': 'value' })
    """
    def __init__(self, metric, tags={}):
        self.metric = metric
        self.tags = tags

    def __call__(self, f):
        def wrapped_f(*args):
            with statsd.timed(self.metric, tags=self.tags):
                return f(*args)
        return wrapped_f


class timer(object):
    """
    with timer('metric.metric', tags={ 'key': 'value' }):
        function()
    """
    def __init__(self, metric, tags={}):
        self.metric = metric
        self.start = None
        self.tags = tags

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        # Must convert to milliseconds, see details in
        # https://statsd.readthedocs.io/en/v3.1/timing.html
        dt = int((time.time() - self.start) * 1000)
        return timing(self.metric, dt, tags=self.tags)



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.shared.utils import wrap_column_name
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)
from typing import Union, Dict

REMOVE_OUTLIERS_TITLE = 'Remove outliers'


class RemoveOutliers(BaseRule):
    default_config = dict(
        max_z_score=3,
    )

    """
    Checks dataframe for the existence of potential outliers, and generates a cleaning suggestion
    to remove these outliers
    """

    def one_dim_outlier_check(self, column: str) -> Union[Dict, None]:
        """
        Checks univariate data for outliers by identifying data that lies more than 3 standard
        deviations outside the mean.

        Args:
            column (str): Column name of the univariate data

        Returns:
            Union[Dict, None]: If a suggestion is made, returns the suggestion dictionary.
            Else returns `None`
        """
        outlier_count = self.statistics.get(f'{column}/outlier_count')
        if outlier_count:
            std = self.statistics[f'{column}/std']
            avg = self.statistics[f'{column}/average']
            max_z_score = self.config('max_z_score')
            upper = avg + max_z_score * std
            lower = avg - max_z_score * std
            wrapped_c = wrap_column_name(column)
            return self._build_transformer_action_suggestion(
                REMOVE_OUTLIERS_TITLE,
                f'Remove {outlier_count} outlier(s) and null values to reduce the amount of noise '
                'in this column.',
                ActionType.FILTER,
                action_arguments=[column],
                action_code=f'{wrapped_c} <= {upper} and {wrapped_c} >= {lower}',
                axis=Axis.ROW,
            )
        return None

    def evaluate(self):
        suggestions = []
        for col in self.df_columns:
            suggestion = self.one_dim_outlier_check(col)
            if suggestion:
                suggestions.append(suggestion)
        return suggestions
